#!/usr/bin/env python
'''
menu handling widgets for wx

Andrew Tridgell
November 2013
'''

from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc
import platform

class MPMenuGeneric(object):
    '''a MP menu separator'''
    def __init__(self):
        pass

    def find_selected(self, event):
        return None

    def _append(self, menu):
        '''append this menu item to a menu'''
        pass

    def __str__(self):
        return "MPMenuGeneric()"

    def __repr__(self):
        return str(self.__str__())

class MPMenuSeparator(MPMenuGeneric):
    '''a MP menu separator'''
    def __init__(self):
        MPMenuGeneric.__init__(self)

    def _append(self, menu):
        '''append this menu item to a menu'''
        menu.AppendSeparator()

    def __str__(self):
        return "MPMenuSeparator()"


class MPMenuItem(MPMenuGeneric):
    '''a MP menu item'''
    def __init__(self, name, description='', returnkey=None, handler=None):
        MPMenuGeneric.__init__(self)
        self.name = name
        self.description = description
        self.returnkey = returnkey
        self.handler = handler
        self.handler_result = None

    def find_selected(self, event):
        '''find the selected menu item'''
        if event.GetId() == self.id():
            return self
        return None

    def call_handler(self):
        '''optionally call a handler function'''
        if self.handler is None:
            return
        call = getattr(self.handler, 'call', None)
        if call is not None:
            self.handler_result = call()

    def id(self):
        '''id used to identify the returned menu items
        uses a 16 bit signed integer'''
        # must be below SHRT_MAX
        id = int(hash((self.name, self.returnkey))) % 32767
        return id

    def _append(self, menu):
        '''append this menu item to a menu'''
        if not self.name:
            return
        menu.Append(self.id(), self.name, self.description)

    def __str__(self):
        return "MPMenuItem(%s,%s,%s)" % (self.name, self.description, self.returnkey)


class MPMenuCheckbox(MPMenuItem):
    '''a MP menu item as a checkbox'''
    def __init__(self, name, description='', returnkey=None, checked=False, handler=None):
        MPMenuItem.__init__(self, name, description=description, returnkey=returnkey, handler=handler)
        self.checked = checked

    def find_selected(self, event):
        '''find the selected menu item'''
        if event.GetId() == self.id():
            self.checked = event.IsChecked()
            return self
        return None

    def IsChecked(self):
        '''return true if item is checked'''
        return self.checked

    def _append(self, menu):
        '''append this menu item to a menu'''
        try:
            menu.AppendCheckItem(self.id(), self.name, self.description)
            menu.Check(self.id(), self.checked)
        except Exception:
            pass

    def __str__(self):
        return "MPMenuCheckbox(%s,%s,%s,%s)" % (self.name, self.description, self.returnkey, str(self.checked))

class MPMenuRadio(MPMenuItem):
    '''a MP menu item as a radio item'''
    def __init__(self, name, description='', returnkey=None, selected=None, items=[], handler=None):
        MPMenuItem.__init__(self, name, description=description, returnkey=returnkey, handler=handler)
        self.items = items
        self.choice = 0
        self.initial = selected

    def set_choices(self, items):
        '''set radio item choices'''
        self.items = items

    def get_choice(self):
        '''return the chosen item'''
        return self.items[self.choice]

    def find_selected(self, event):
        '''find the selected menu item'''
        first = self.id()
        last = first + len(self.items) - 1
        evid = event.GetId()
        if evid >= first and evid <= last:
            self.choice = evid - first
            return self
        return None

    def _append(self, menu):
        '''append this menu item to a menu'''
        from MAVProxy.modules.lib.wx_loader import wx
        submenu = wx.Menu()
        for i in range(len(self.items)):
            submenu.AppendRadioItem(self.id()+i, self.items[i], self.description)
            if self.items[i] == self.initial:
                submenu.Check(self.id()+i, True)
        menu.AppendSubMenu(submenu, self.name)

    def __str__(self):
        return "MPMenuRadio(%s,%s,%s,%s)" % (self.name, self.description, self.returnkey, self.get_choice())


class MPMenuSubMenu(MPMenuGeneric):
    '''a MP menu item'''
    def __init__(self, name, items):
        MPMenuGeneric.__init__(self)
        self.name = name
        self.items = items

    def add(self, items, addto=None):
        '''add more items to a sub-menu'''
        if not isinstance(items, list):
            items = [items]
        for m in items:
            updated = False
            for i in range(len(self.items)):
                try:
                    if self.items[i].name == m.name:
                        self.items[i] = m
                        updated = True
                except Exception:
                    pass
            if not updated:
                self.items.append(m)

    def remove(self, items):
        '''remove items from a sub-menu'''
        if not isinstance(items, list):
            items = [items]
        names = set([x.name for x in items])
        self.items = list(filter(lambda x : x.name not in names, self.items))

    def add_to_submenu(self, submenu_path, item):
        '''add an item to a submenu using a menu path array'''
        if len(submenu_path) == 0:
            self.add(item)
            return
        for m in self.items:
            if isinstance(m, MPMenuSubMenu) and submenu_path[0] == m.name:
                m.add_to_submenu(submenu_path[1:], item)
                return
        self.add(MPMenuSubMenu(submenu_path[0], []))
        self.add_to_submenu(submenu_path, item)

    def combine(self, submenu):
        '''combine a new menu with an existing one'''
        self.items.extend(submenu.items)

    def wx_menu(self):
        '''return a wx.Menu() for this menu'''
        from MAVProxy.modules.lib.wx_loader import wx
        menu = wx.Menu()
        for i in range(len(self.items)):
            m = self.items[i]
            m._append(menu)
        return menu

    def find_selected(self, event):
        '''find the selected menu item'''
        for m in self.items:
            ret = m.find_selected(event)
            if ret is not None:
                return ret
        return None

    def _append(self, menu):
        '''append this menu item to a menu'''
        menu.AppendSubMenu(self.wx_menu(), self.name) #use wxPython_phoenix

    def __str__(self):
        return "MPMenuSubMenu(%s)" % (self.name)


class MPMenuTop(object):
    '''a MP top level menu'''
    def __init__(self, items):
        self.items = items

    def add(self, items):
        '''add a submenu'''
        if not isinstance(items, list):
            items = [items]
        for m in items:
            updated = False
            for i in range(len(self.items)):
                if self.items[i].name == m.name:
                    self.items[i] = m
                    updated = True
            if not updated:
                self.items.append(m)

    def remove(self, items):
        if not isinstance(items, list):
            items = [items]
        names = set([x.name for x in items])
        self.items = list(filter(lambda x : x.name not in names, self.items))

    def add_to_submenu(self, submenu_path, item):
        '''
        add an item to a submenu using a menu path array
        '''
        for m in self.items:
            if m.name == submenu_path[0]:
                m.add_to_submenu(submenu_path[1:], item)
                return
        raise(ValueError("No submenu (%s) found" % (submenu_path[0])))

    def wx_menu(self):
        '''return a wx.MenuBar() for the menu'''
        from MAVProxy.modules.lib.wx_loader import wx

        menubar = wx.MenuBar()
        for i in range(len(self.items)):
            m = self.items[i]
            menubar.Append(m.wx_menu(), m.name)
        return menubar

    def find_selected(self, event):
        '''find the selected menu item'''
        for i in range(len(self.items)):
            m = self.items[i]
            ret = m.find_selected(event)
            if ret is not None:
                return ret
        return None

class MPMenuCallFileDialog(object):
    '''used to create a file dialog callback'''
    def __init__(self, flags=None, title='Filename', wildcard='*.*'):
        self.flags = flags or ('open',)
        self.title = title
        self.wildcard = wildcard

    def call(self):
        '''show a file dialog'''
        from MAVProxy.modules.lib.wx_loader import wx

        # remap flags to wx descriptors
        flag_map = {
            'open': wx.FD_OPEN,
            'save': wx.FD_SAVE,
            'overwrite_prompt': wx.FD_OVERWRITE_PROMPT,
        }
        flagsMapped = list(map(lambda x: flag_map[x], self.flags))

        #need to OR together the elements of the flagsMapped tuple
        if len(flagsMapped) == 1:
            dlg = wx.FileDialog(None, self.title, '', "", self.wildcard, flagsMapped[0])
        else:
            dlg = wx.FileDialog(None, self.title, '', "", self.wildcard, flagsMapped[0]|flagsMapped[1])
        if dlg.ShowModal() != wx.ID_OK:
            return None
        return "\"" + dlg.GetPath() + "\""

class MPMenuCallTextDialog(object):
    '''used to create a value dialog callback'''
    def __init__(self, title='Enter Value', default=''):
        self.title = title
        self.default = default

    def call(self):
        '''show a value dialog'''
        from MAVProxy.modules.lib.wx_loader import wx
        try:
            dlg = wx.TextEntryDialog(None, self.title, self.title, defaultValue=str(self.default))
        except TypeError:
            dlg = wx.TextEntryDialog(None, self.title, self.title, value=str(self.default))
        if dlg.ShowModal() != wx.ID_OK:
            return None
        return dlg.GetValue()

class MPMenuChildMessageDialog(object):
    '''used to create a message dialog in a child process'''
    def __init__(self, title='Information', message='', font_size=18):
        self.title = title
        self.message = message
        self.font_size = font_size

    def show(self):
        t = multiproc.Process(target=self.call)
        t.start()

    def call(self):
        '''show the dialog as a child process'''
        mp_util.child_close_fds()
        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx
        from wx.lib.agw.genericmessagedialog import GenericMessageDialog
        app = wx.App(False)
        # note! font size change is not working. I don't know why yet
        font = wx.Font(self.font_size, wx.MODERN, wx.NORMAL, wx.NORMAL)
        dlg = GenericMessageDialog(None, self.message, self.title, wx.ICON_INFORMATION|wx.OK)
        dlg.SetFont(font)
        dlg.ShowModal()
        app.MainLoop()

class MPMenuOpenWeblink(object):
    '''used to open a weblink in the default webbrowser'''
    def __init__(self, url='www.google.com'):
        self.url = url

    def call(self):
        '''show the dialog as a child process'''
        import webbrowser
        webbrowser.open_new_tab(self.url)

if __name__ == '__main__':
    from MAVProxy.modules.lib.mp_image import MPImage
    import time
    im = MPImage(mouse_events=True,
                 key_events=True,
                 can_drag = False,
                 can_zoom = False,
                 auto_size = True)

    menu = MPMenuTop([MPMenuSubMenu('&File',
                                    items=[MPMenuItem('&Open\tCtrl+O'),
                                           MPMenuItem('&Save\tCtrl+S'),
                                           MPMenuItem('Close', 'Close'),
                                           MPMenuItem('&Quit\tCtrl+Q', 'Quit')]),
                      MPMenuSubMenu('Edit',
                                    items=[MPMenuSubMenu('Option',
                                                         items=[MPMenuItem('Foo'),
                                                                MPMenuItem('Bar'),
                                                                MPMenuSeparator(),
                                                                MPMenuCheckbox('&Grid\tCtrl+G')]),
                                           MPMenuItem('Image', 'EditImage'),
                                           MPMenuRadio('Colours',
                                                       items=['Red','Green','Blue']),
                                           MPMenuRadio('Shapes',
                                                       items=['Circle','Square','Triangle'])])])

    im.set_menu(menu)

    popup = MPMenuSubMenu('A Popup',
                          items=[MPMenuItem('Sub1'),
                                 MPMenuItem('Sub2'),
                                 MPMenuItem('Sub3')])

    im.set_popup_menu(popup)

    while im.is_alive():
        for event in im.events():
            if isinstance(event, MPMenuItem):
                print(event, getattr(event, 'popup_pos', None))
                continue
            else:
                print(event)
        time.sleep(0.1)



#!/usr/bin/env python
'''remote low level device operations'''

import time, os, sys
from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module

class DeviceOpModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(DeviceOpModule, self).__init__(mpstate, "DeviceOp")
        self.add_command('devop', self.cmd_devop, "device operations",
                         ["<read|write> <spi|i2c>"])
        self.request_id = 1

        self.failure_strings = {
            1: "No such bus type",
            2: "No such bus/device",
            3: "Sempahore-take failure",
            4: "transfer failed",
        }

    def cmd_devop(self, args):
        '''device operations'''
        usage = "Usage: devop <read|write> <spi|i2c> name bus address"
        if len(args) < 5:
            print(usage)
            return

        if args[1] == 'spi':
            bustype = mavutil.mavlink.DEVICE_OP_BUSTYPE_SPI
        elif args[1] == 'i2c':
            bustype = mavutil.mavlink.DEVICE_OP_BUSTYPE_I2C
        else:
            print(usage)

        if args[0] == 'read':
            self.devop_read(args[2:], bustype)
        elif args[0] == 'write':
            self.devop_write(args[2:], bustype)
        else:
            print(usage)

    def devop_read(self, args, bustype):
        '''read from device'''
        if len(args) < 5:
            print("Usage: devop read <spi|i2c> name bus address regstart count")
            return
        name = args[0]
        bus = int(args[1],base=0)
        address = int(args[2],base=0)
        reg = int(args[3],base=0)
        count = int(args[4],base=0)
        if sys.version_info.major >= 3:
            name = bytearray(name, 'ascii')
        self.master.mav.device_op_read_send(self.target_system,
                                            self.target_component,
                                            self.request_id,
                                            bustype,
                                            bus,
                                            address,
                                            name,
                                            reg,
                                            count)
        self.request_id += 1

    def devop_write(self, args, bustype):
        '''write to a device'''
        usage = "Usage: devop write <spi|i2c> name bus address regstart count <bytes>"
        if len(args) < 5:
            print(usage)
            return
        name = args[0]
        bus = int(args[1],base=0)
        address = int(args[2],base=0)
        reg = int(args[3],base=0)
        count = int(args[4],base=0)
        args = args[5:]
        if len(args) < count:
            print(usage)
            return
        bytes = [0]*128
        for i in range(count):
            bytes[i] = int(args[i],base=0)
        if sys.version_info.major >= 3:
            name = bytearray(name, 'ascii')
        self.master.mav.device_op_write_send(self.target_system,
                                             self.target_component,
                                             self.request_id,
                                             bustype,
                                             bus,
                                             address,
                                             name,
                                             reg,
                                             count,
                                             bytes)
        self.request_id += 1


    def read_show_reply(self, m):
        # header
        sys.stdout.write("     ")
        for i in range(16):
            sys.stdout.write("%3x" % i)
        print("")
        sys.stdout.write("%4x: " % (m.regstart-m.regstart%16,))
        # leading no-data-read:
        for i in range(m.regstart%16):
            sys.stdout.write("-- ")

        for i in range(m.count):
            reg = i + m.regstart
            sys.stdout.write("%02x " % (m.data[i]))
            if (reg+1) % 16 == 0 and i != m.count-1:
                print("")
                sys.stdout.write("%4x: " % ((reg+1)-(reg+1)%16,))

        # trailing no-data-read
        if (m.regstart+m.count) % 16 != 0:
            if m.count < 16 and m.regstart % 16 != 0 and (m.regstart%16+m.count) <16:
                # front of line is padded
                count = 16 - m.regstart%16 - m.count
            else:
                count = 16 - (m.regstart+m.count)%16

            for i in range(0,count):
                sys.stdout.write("-- ")
        print("")

    def mavlink_packet(self, m):
        '''handle a mavlink packet'''
        mtype = m.get_type()
        if mtype == "DEVICE_OP_READ_REPLY":
            if m.result != 0:
                print("Operation %u failed: %u (%s)" %
                      (m.request_id,
                       m.result,
                       self.failure_strings.get(m.result, '????')))
            else:
                print("Operation %u OK: %u bytes" % (m.request_id, m.count))
                self.read_show_reply(m)

        if mtype == "DEVICE_OP_WRITE_REPLY":
            if m.result != 0:
                print("Operation %u failed: %u (%s)" %
                      (m.request_id,
                       m.result,
                       self.failure_strings.get(m.result, '????')))
            else:
                print("Operation %u OK" % m.request_id)

def init(mpstate):
    '''initialise module'''
    return DeviceOpModule(mpstate)



#!/usr/bin/env python
'''
support for a GCS attached DGPS system
'''

import socket, errno
from pymavlink import mavutil
from MAVProxy.modules.lib import mp_module

class DGPSModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(DGPSModule, self).__init__(mpstate, "DGPS", "DGPS injection support for SBP/RTCP/UBC")
        self.portnum = 13320
        self.port = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.port.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.port.bind(("127.0.0.1", self.portnum))
        mavutil.set_close_on_exec(self.port.fileno())
        self.port.setblocking(0)
        self.inject_seq_nr = 0
        print("DGPS: Listening for RTCM packets on UDP://%s:%s" % ("127.0.0.1", self.portnum))
    
    def send_rtcm_msg(self, data):
        msglen = 180;
        
        if (len(data) > msglen * 4):
            print("DGPS: Message too large", len(data))
            return
        
        # How many messages will we send?
        msgs = 0
        if (len(data) % msglen == 0):
            msgs = len(data) // msglen
        else:
            msgs = (len(data) // msglen) + 1

        for a in range(0, msgs):
            
            flags = 0
            
            # Set the fragment flag if we're sending more than 1 packet.
            if (msgs) > 1:
                flags = 1
            
            # Set the ID of this fragment
            flags |= (a & 0x3) << 1
            
            # Set an overall sequence number
            flags |= (self.inject_seq_nr & 0x1f) << 3
            
            
            amount = min(len(data) - a * msglen, msglen)
            datachunk = data[a*msglen : a*msglen + amount]

            self.master.mav.gps_rtcm_data_send(
                flags,
                len(datachunk),
                bytearray(datachunk.ljust(180, b'\0')))
        
        # Send a terminal 0-length message if we sent 2 or 3 exactly-full messages.     
        if (msgs < 4) and (len(data) % msglen == 0) and (len(data) > msglen):
            flags = 1 | (msgs & 0x3)  << 1 | (self.inject_seq_nr & 0x1f) << 3
            self.master.mav.gps_rtcm_data_send(
                flags,
                0,
                bytearray("".ljust(180, '\0')))
            
        self.inject_seq_nr += 1



    def idle_task(self):
        '''called in idle time'''
        try:
            data = self.port.recv(1024) # Attempt to read up to 1024 bytes.
        except socket.error as e:
            if e.errno in [ errno.EAGAIN, errno.EWOULDBLOCK ]:
                return
            raise
        try:
            self.send_rtcm_msg(data)

        except Exception as e:
            print("DGPS: GPS Inject Failed:", e)

def init(mpstate):
    '''initialise module'''
    return DGPSModule(mpstate)




#!/usr/bin/env python
'''remote nsh console handling'''

import time, os, fnmatch, sys, time
from pymavlink import mavutil, mavwp
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import mp_module

class NSHModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(NSHModule, self).__init__(mpstate, "nsh", "remote nsh shell")
        self.add_command('nsh', self.cmd_nsh,
                         'nsh shell control',
                         ['<start|stop>',
                          'set (SERIALSETTING)'])
        self.serial_settings = mp_settings.MPSettings(
            [ ('port', int, mavutil.mavlink.SERIAL_CONTROL_DEV_SHELL),
              ('baudrate', int, 57600)
              ]
            )
        self.add_completion_function('(SERIALSETTING)', self.serial_settings.completion)
        self.last_packet = time.time()
        self.last_check = time.time()
        self.started = False

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'SERIAL_CONTROL':
            data = m.data[:m.count]
            if m.count > 0:
                s = ''.join(str(chr(x)) for x in data)
                if self.mpstate.system == 'Windows':
                    # strip nsh ansi codes
                    s = s.replace("\033[K","")
                sys.stdout.write(s)
                self.last_packet = time.time()

    def stop(self):
        '''stop nsh input'''
        self.set_prompt(self.status.flightmode + "> ")
        self.mpstate.functions.input_handler = None
        self.started = False
        # unlock the port
        mav = self.master.mav
        mav.serial_control_send(self.serial_settings.port,
                                0,
                                0, self.serial_settings.baudrate,
                                0, [0]*70)

    def send(self, line):
        '''send some bytes'''
        line = line.strip()
        if line == ".":
            self.stop()
            return
        mav = self.master.mav
        if line != '+++':
            line += "\r\n"
        buf = [ord(x) for x in line]
        buf.extend([0]*(70-len(buf)))

        flags = mavutil.mavlink.SERIAL_CONTROL_FLAG_RESPOND
        flags |= mavutil.mavlink.SERIAL_CONTROL_FLAG_MULTI
        flags |= mavutil.mavlink.SERIAL_CONTROL_FLAG_EXCLUSIVE
        mav.serial_control_send(self.serial_settings.port,
                                flags,
                                0, self.serial_settings.baudrate,
                                len(line), buf)

    def idle_task(self):
        '''handle mavlink packets'''
        if not self.started:
            return
        now = time.time()
        if now - self.last_packet < 1:
            timeout = 0.05
        else:
            timeout = 0.2
        if now - self.last_check > timeout:
            self.last_check = now
            mav = self.master.mav
            flags = mavutil.mavlink.SERIAL_CONTROL_FLAG_RESPOND
            flags |= mavutil.mavlink.SERIAL_CONTROL_FLAG_MULTI
            flags |= mavutil.mavlink.SERIAL_CONTROL_FLAG_EXCLUSIVE
            mav.serial_control_send(self.serial_settings.port,
                                    flags,
                                    0, self.serial_settings.baudrate,
                                    0, [0]*70)


    def cmd_nsh(self, args):
        '''nsh shell commands'''
        usage = "Usage: nsh <start|stop|set>"
        if len(args) < 1:
            print(usage)
            return
        if args[0] == "start":
            self.mpstate.functions.input_handler = self.send
            self.started = True
            self.mpstate.rl.set_prompt("")
        elif args[0] == "stop":
            self.stop()
        elif args[0] == "set":
            self.serial_settings.command(args[1:])
        else:
            print(usage)

def init(mpstate):
    '''initialise module'''
    return NSHModule(mpstate)



from MAVProxy.modules.lib import wx_util

if not wx_util.safe:
    print('Cannot access wx from main thread.')
    import traceback
    print(traceback.print_stack())
    raise Exception('Cannot access wx from main thread')

import wx



from __future__ import print_function

import os
import pygame
import pkg_resources
import yaml
import fnmatch

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings

from MAVProxy.modules.mavproxy_joystick import controls


class Joystick(mp_module.MPModule):
    '''
    joystick set verbose
    joystick set debug
    joystick status
    joystick probe
    '''

    def __init__(self, mpstate):
        """Initialise module"""
        super(Joystick, self).__init__(mpstate, 'joystick',
                                       'A flexible joystick driver')

        self.joystick = None

        self.init_pygame()
        self.init_settings()
        self.init_commands()

        self.probe()

    def log(self, msg, level=0):
        if self.mpstate.settings.moddebug < level:
            return

        print('{}: {}'.format(__name__, msg))

    def init_pygame(self):
        self.log('Initializing pygame', 2)
        pygame.init()
        pygame.joystick.init()

    def init_settings(self):
        pass

    def init_commands(self):
        self.log('Initializing commands', 2)
        self.add_command('joystick', self.cmd_joystick,
                         "A flexible joystick drvier",
                         ['status',  'probe'])

    def load_definitions(self):
        self.log('Loading joystick definitions', 1)

        self.joydefs = []
        search = []

        userjoysticks = os.environ.get(
            'MAVPROXY_JOYSTICK_DIR',
            mp_util.dot_mavproxy('joysticks'))
        if userjoysticks is not None and os.path.isdir(userjoysticks):
            search.append(userjoysticks)

        search.append(pkg_resources.resource_filename(__name__, 'joysticks'))

        for path in search:
            self.log('Looking for joystick definitions in {}'.format(path),
                     2)
            path = os.path.expanduser(path)
            for dirpath, dirnames, filenames in os.walk(path):
                for joyfile in filenames:
                    root, ext = os.path.splitext(joyfile)
                    if ext[1:] not in ['yml', 'yaml', 'json']:
                        continue

                    joypath = os.path.join(dirpath, joyfile)
                    self.log('Loading definition from {}'.format(joypath), 2)
                    with open(joypath, 'r') as fd:
                        joydef = yaml.safe_load(fd)
                        joydef['path'] = joypath
                        self.joydefs.append(joydef)

    def probe(self):
        self.load_definitions()

        for jid in range(pygame.joystick.get_count()):
            joy = pygame.joystick.Joystick(jid)
            self.log("Found joystick (%s)" % (joy.get_name(),))
            for joydef in self.joydefs:
                if 'match' not in joydef:
                    self.log('{} has no match patterns, ignoring.'.format(
                        joydef['path']), 0)
                    continue
                for pattern in joydef['match']:
                    if fnmatch.fnmatch(joy.get_name().lower(),
                                       pattern.lower()):
                        self.log('Using {} ("{}" matches pattern "{}")'.format(
                            joydef['path'], joy.get_name(), pattern))
                        self.joystick = controls.Joystick(joy, joydef)
                        return

        print('{}: Failed to find matching joystick.'.format(__name__))

    def usage(self):
        '''show help on command line options'''
        return "Usage: joystick <status|set>"

    def cmd_joystick(self, args):
        if not len(args):
            self.log('No subcommand specified.')
        elif args[0] == 'status':
            self.cmd_status()
        elif args[0] == 'probe':
            self.cmd_probe()
        elif args[0] == 'help':
            self.cmd_help()

    def cmd_help(self):
        print('joystick probe -- reload and match joystick definitions')
        print('joystick status -- show currently loaded definition, if any')

    def cmd_probe(self):
        self.log('Re-detecting available joysticks', 0)
        self.probe()

    def cmd_status(self):
        if self.joystick is None:
            print('No active joystick')
        else:
            print('Active joystick:')
            print('Path: {path}'.format(**self.joystick.controls))
            print('Description: {description}'.format(
                **self.joystick.controls))

    def idle_task(self):
        if self.joystick is None:
            return

        for e in pygame.event.get():
            override = self.module('rc').override[:]
            values = self.joystick.read()
            override = values + override[len(values):]

            # self.log('channels: {}'.format(override), level=3)

            if override != self.module('rc').override:
                self.module('rc').override = override
                self.module('rc').override_period.force()


def init(mpstate):
    '''initialise module'''
    return Joystick(mpstate)



import time

class MPModule(object):
    '''
    The base class for all modules
    '''

    def __init__(self, mpstate, name, description=None, public=False, multi_instance=False, multi_vehicle=False):
        '''
        Constructor

        if public is true other modules can find this module instance with module('name')
        '''
        self.mpstate = mpstate
        self.name = name
        self.needs_unloading = False
        self.multi_instance = multi_instance
        self.multi_vehicle = multi_vehicle

        if description is None:
            self.description = name + " handling"
        else:
            self.description = description
        if multi_instance:
            if not name in mpstate.multi_instance:
                mpstate.multi_instance[name] = []
                mpstate.instance_count[name] = 0
            mpstate.multi_instance[name].append(self)
            mpstate.instance_count[name] += 1
            self.instance = mpstate.instance_count[name]
            if self.instance > 1:
                # make the name distincitive, so self.module('map2') works
                self.name += str(self.instance)
        if public:
            mpstate.public_modules[self.name] = self

    #
    # Overridable hooks follow...
    #

    def idle_task(self):
        pass

    def unload(self):
        if self.multi_instance and self.name in self.mpstate.multi_instance:
            self.mpstate.multi_instance.pop(self.name)

    def unknown_command(self, args):
        '''Return True if we have handled the unknown command'''
        return False

    def mavlink_packet(self, packet):
        pass

    #
    # Methods for subclass use
    #

    def module(self, name):
        '''Find a public module (most modules are private)'''
        return self.mpstate.module(name)

    def module_matching(self, name):
        '''Find a list of modules matching a wildcard pattern'''
        import fnmatch
        ret = []
        for mname in self.mpstate.public_modules.keys():
            if fnmatch.fnmatch(mname, name):
                ret.append(self.mpstate.public_modules[mname])
        return ret

    def get_time(self):
        '''get time, using ATTITUDE.time_boot_ms if in SITL with SIM_SPEEDUP != 1'''
        systime = time.time() - self.mpstate.start_time_s
        if not self.mpstate.is_sitl:
            return systime
        try:
            speedup = int(self.get_mav_param('SIM_SPEEDUP',1))
        except Exception:
            return systime
        if speedup != 1:
            return self.mpstate.attitude_time_s
        return systime

    @property
    def console(self):
        return self.mpstate.console

    @property
    def status(self):
        return self.mpstate.status

    @property
    def mav_param(self):
        return self.mpstate.mav_param

    @property
    def settings(self):
        return self.mpstate.settings

    @property
    def vehicle_type(self):
        return self.mpstate.vehicle_type

    @property
    def vehicle_name(self):
        return self.mpstate.vehicle_name

    @property
    def sitl_output(self):
        return self.mpstate.sitl_output

    @property
    def target_system(self):
        return self.settings.target_system

    @property
    def target_component(self):
        return self.settings.target_component

    @property
    def master(self):
        return self.mpstate.master()

    @property
    def continue_mode(self):
        return self.mpstate.continue_mode

    @property
    def logdir(self):
        return self.mpstate.status.logdir

    def say(self, msg, priority='important'):
        return self.mpstate.functions.say(msg)

    def get_mav_param(self, param_name, default=None):
        return self.mpstate.functions.get_mav_param(param_name, default)

    def param_set(self, name, value, retries=3):
        self.mpstate.functions.param_set(name, value, retries)

    def add_command(self, name, callback, description, completions=None):
        self.mpstate.command_map[name] = (callback, description)
        if completions is not None:
            self.mpstate.completions[name] = completions

    def remove_command(self, name):
        if name in self.mpstate.command_map:
            del self.mpstate.command_map[name]
        if name in self.mpstate.completions:
            del self.mpstate.completions[name]

    def add_completion_function(self, name, callback):
        self.mpstate.completion_functions[name] = callback

    def dist_string(self, val_meters):
        '''return a distance as a string'''
        if self.settings.dist_unit == 'nm':
            return "%.1fnm" % (val_meters * 0.000539957)
        if self.settings.dist_unit == 'miles':
            return "%.1fmiles" % (val_meters * 0.000621371)
        return "%um" % val_meters

    def height_convert_units(self, val_meters):
        '''return a height in configured units'''
        if self.settings.height_unit == 'feet':
            return val_meters * 3.28084
        return val_meters

    def height_string(self, val_meters):
        '''return a height as a string'''
        if self.settings.height_unit == 'feet':
            return "%uft" % (val_meters * 3.28084)
        return "%um" % val_meters

    def speed_convert_units(self, val_ms):
        '''return a speed in configured units'''
        if self.settings.speed_unit == 'knots':
            return val_ms * 1.94384
        elif self.settings.speed_unit == 'mph':
            return val_ms * 2.23694
        return val_ms

    def speed_string(self, val_ms):
        '''return a speed as a string'''
        if self.settings.speed_unit == 'knots':
            return "%ukn" % (val_ms * 1.94384)
        elif self.settings.speed_unit == 'mph':
            return "%umph" % (val_ms * 2.23694)
        return "%um/s" % val_ms

    def set_prompt(self, prompt):
        '''set prompt for command line'''
        if prompt and self.settings.vehicle_name:
            # add in optional vehicle name
            prompt = self.settings.vehicle_name + ':' + prompt
        self.mpstate.rl.set_prompt(prompt)
            
    @staticmethod
    def link_label(link):
        '''return a link label as a string'''
        if hasattr(link, 'label'):
            label = link.label
        else:
            label = str(link.linknum+1)
        return label

    def is_primary_vehicle(self, msg):
        '''see if a msg is from our primary vehicle'''
        sysid = msg.get_srcSystem()
        if self.target_system == 0 or self.target_system == sysid:
            return True
        return False



#!/usr/bin/env python
'''
nmea serial output module
Matthew Ridley
August 2012

UAV outback challenge search and rescue rules
5.15 Situational Awareness Requirement

It is highly desirable that teams provide:
 - an NMEA 0183 serial output with GPRMC and GPGGA sentences for
   aircraft current location

'''

import sys, os, serial
from MAVProxy.modules.lib import mp_module

class NMEAModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(NMEAModule, self).__init__(mpstate, "NMEA", "NMEA output")
        self.port = None
        self.baudrate = 4800
        self.data = 8
        self.parity = 'N'
        self.stop = 1
        self.serial = None
        self.output_time = 0.0
        self.add_command('nmea', self.cmd_nmea, "nmea control")

    def cmd_nmea(self, args):
        '''set nmea'''
        usage = "nmea port [baudrate data parity stop]"
        if len(args) == 0:
            if self.port is None:
                print("NMEA output port not set")
                print(usage)
            else:
                print("NMEA output port %s, %d, %d, %s, %d" % (str(self.port), self.baudrate, self.data, str(self.parity), self.stop))
            return
        if len(args) > 0:
            self.port = str(args[0])
        if len(args) > 1:
            self.baudrate = int(args[1])
        if len(args) > 2:
            self.data = int(args[2])
        if len(args) > 3:
            self.parity = str(args[3])
        if len(args) > 4:
            self.stop = int(args[4])

        if self.serial is not None:
            self.serial.close()
        self.serial = None
        if len(args) > 0:
            if self.port.startswith("/dev/"):
                try:
                    self.serial = serial.Serial(self.port, self.baudrate, self.data, self.parity, self.stop)
                except serial.SerialException as se:
                    print("Failed to open output port %s:%s" % (self.port, se.message))
            else:
                self.serial = open(self.port, mode='w')
            

    def format_date(self, utc_sec):
        import time
        tm_t = time.gmtime(utc_sec)
        return "%02d%02d%02d" % (tm_t.tm_mday, tm_t.tm_mon, tm_t.tm_year % 100)

    def format_time(self, utc_sec):
        import time
        tm_t = time.gmtime(utc_sec)
        subsecs = utc_sec - int(utc_sec);
        return "%02d%02d%05.3f" % (tm_t.tm_hour, tm_t.tm_min, tm_t.tm_sec + subsecs)

    def format_lat(self, lat):
        deg = abs(lat)
        minutes = (deg - int(deg))*60
        return "%02d%08.5f,%c" % (int(deg), minutes, 'S' if lat < 0 else 'N')

    def format_lon(self, lon):
        deg = abs(lon)
        minutes = (deg - int(deg))*60
        return "%03d%08.5f,%c" % (int(deg), minutes, 'W' if lon < 0 else 'E')

	# tst = "$GPRMC,225446,A,4916.45,N,12311.12,W,000.5,054.7,191194,020.3,E*68"
	# print ("*%02X" % nmea_checksum(tst))

    def nmea_checksum(self, msg):
        d = msg[1:]
        cs = 0
        for i in d:
            cs ^= ord(i)
        return cs

    def format_gga(self, utc_sec, lat, lon, fix, nsat, hdop, alt):
        fmt = "$GPGGA,%s,%s,%s,%01d,%02d,%04.1f,%07.2f,M,0.0,M,,"
        msg = fmt % (self.format_time(utc_sec), self.format_lat(lat), self.format_lon(lon), fix, nsat, hdop, alt)
        return msg + "*%02X" % self.nmea_checksum(msg)

    def format_rmc(self, utc_sec, fix, lat, lon, speed, course):
        fmt = "$GPRMC,%s,%s,%s,%s,%.2f,%.2f,%s,,"
        msg = fmt % (self.format_time(utc_sec), fix, self.format_lat(lat), self.format_lon(lon),
                     speed, course, self.format_date(utc_sec))
        return msg + "*%02X" % self.nmea_checksum(msg)

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        import time

        now_time = time.time()
        if abs(self.output_time - now_time) < 1.0:
            return

        if m.get_type() == 'GPS_RAW_INT':
            # for GPRMC and GPGGA
            utc_sec = now_time
            fix_status = 'A' if (m.fix_type > 1) else 'V'
            lat = m.lat/1.0e7
            lon = m.lon/1.0e7

            # for GPRMC
            knots = ((m.vel/100.0)/1852.0)*3600
            course = m.cog/100.0

            # for GPGGA
            fix_quality = 1 if (m.fix_type > 1) else 0 # 0/1 for (in)valid or 2 DGPS
            num_sat = m.satellites_visible
            hdop = m.eph/100.0
            altitude = m.alt/1000.0

            # print format_gga(utc_sec, lat, lon, fix_quality, num_sat, hdop, altitude)
            # print format_rmc(utc_sec, fix_status, lat, lon, knots, course)
            gga = self.format_gga(utc_sec, lat, lon, fix_quality, num_sat, hdop, altitude)
            rmc = self.format_rmc(utc_sec, fix_status, lat, lon, knots, course)

            self.output_time = now_time
            #print(gga+'\r')
            #print(rmc+'\r')
            #print(self.serial)
            if self.serial is not None:
                self.serial.write(gga + '\r\n')
                self.serial.write(rmc + '\r\n')
                self.serial.flush()
     

def init(mpstate):
    '''initialise module'''
    return NMEAModule(mpstate)



'''
generate dynamic obstacles for OBC 2018
'''

import time, pickle
from math import *

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import mp_util
from pymavlink import mavutil
if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

import socket, time, random, math

track_count = 0

# object types
DNFZ_types = {
    'Aircraft' : 1,
    'Weather' : 20000,
    'BirdMigrating' : 30000,
    'BirdOfPrey' : 40000
}

from MAVProxy.modules.mavproxy_map import mp_elevation

ElevationMap = mp_elevation.ElevationModel()

gen_settings = mp_settings.MPSettings([("port", int, 45454),
                                       ('debug', int, 0),
                                       ('home_lat', float, -27.298440),
                                       ('home_lon', float, 151.290775),
                                       ('region_width', float, 15000),
                                       ('num_aircraft', int, 5),
                                       ('num_bird_prey', int, 5),
                                       ('num_bird_migratory', int, 5),
                                       ('num_weather', int, 5),
                                       ('wgs84_to_AMSL', float, -41.2),
                                       ('stop', int, 0)])
                                       
    
class DNFZ:
    '''a dynamic no-fly zone object'''
    def __init__(self, DNFZ_type):
        if not DNFZ_type in DNFZ_types:
            raise('Bad DNFZ type %s' % DNFZ_type)
        self.DNFZ_type = DNFZ_type
        self.pkt = {'category': 0, 'I010': {'SAC': {'val': 4, 'desc': 'System Area Code'}, 'SIC': {'val': 0, 'desc': 'System Identification Code'}}, 'I040': {'TrkN': {'val': 0, 'desc': 'Track number'}}, 'ts': 0, 'len': 25, 'I220': {'RoC': {'val': 0.0, 'desc': 'Rate of Climb/Descent'}}, 'crc': 'B52DA163', 'I130': {'Alt': {'max': 150000.0, 'min': -1500.0, 'val': 0.0, 'desc': 'Altitude'}}, 'I070': {'ToT': {'val': 0.0, 'desc': 'Time Of Track Information'}}, 'I105': {'Lat': {'val': 0, 'desc': 'Latitude in WGS.84 in twos complement. Range -90 < latitude < 90 deg.'}, 'Lon': {'val': 0.0, 'desc': 'Longitude in WGS.84 in twos complement. Range -180 < longitude < 180 deg.'}}, 'I080': {'SRC': {'meaning': '3D radar', 'val': 2, 'desc': 'Source of calculated track altitude for I062/130'}, 'FX': {'meaning': 'end of data item', 'val': 0, 'desc': ''}, 'CNF': {'meaning': 'Confirmed track', 'val': 0, 'desc': ''}, 'SPI': {'meaning': 'default value', 'val': 0, 'desc': ''}, 'MRH': {'meaning': 'Geometric altitude more reliable', 'val': 1, 'desc': 'Most Reliable Height'}, 'MON': {'meaning': 'Multisensor track', 'val': 0, 'desc': ''}}}
        self.speed = 0.0 # m/s
        self.heading = 0.0 # degrees
        self.desired_heading = None
        self.yawrate = 0.0
        # random initial position and heading
        self.randpos()
        self.setheading(random.uniform(0,360))
        self.setclimbrate(random.uniform(-3,3))
        global track_count
        track_count += 1
        self.pkt['I040']['TrkN']['val'] = DNFZ_types[self.DNFZ_type] + track_count
        if gen_settings.debug > 0:
            print("track %u" % self.pkt['I040']['TrkN']['val'])

    def distance_from(self, lat, lon):
        '''get distance from a point'''
        lat1 = self.pkt['I105']['Lat']['val']
        lon1 = self.pkt['I105']['Lon']['val']
        return mp_util.gps_distance(lat1, lon1, lat, lon)
        
    def distance_from_home(self):
        '''get distance from home'''
        return self.distance_from(gen_settings.home_lat, gen_settings.home_lon)
        
    def randpos(self):
        '''random initial position'''
        self.setpos(gen_settings.home_lat, gen_settings.home_lon)
        self.move(random.uniform(0, 360), random.uniform(0, gen_settings.region_width))

    def ground_height(self):
        '''return height above ground in feet'''
        lat = self.pkt['I105']['Lat']['val']
        lon = self.pkt['I105']['Lon']['val']
        global ElevationMap
        ret = ElevationMap.GetElevation(lat, lon)
        ret -= gen_settings.wgs84_to_AMSL
        return ret * 3.2807

    def randalt(self):
        '''random initial position'''
        self.setalt(self.ground_height() + random.uniform(100, 1500))

    def move(self, bearing, distance):
        '''move position by bearing and distance'''
        lat = self.pkt['I105']['Lat']['val']
        lon = self.pkt['I105']['Lon']['val']
        (lat, lon) = mp_util.gps_newpos(lat, lon, bearing, distance)
        self.setpos(lat, lon)
        
    def setpos(self, lat, lon):
        self.pkt['I105']['Lat']['val'] = lat
        self.pkt['I105']['Lon']['val'] = lon

    def getlat(self):
        return self.pkt['I105']['Lat']['val']

    def getlon(self):
        return self.pkt['I105']['Lon']['val']

    def getalt(self):
        return self.pkt['I130']['Alt']['val']
        
    def setalt(self, alt):
        self.pkt['I130']['Alt']['val'] = alt

    def setclimbrate(self, climbrate):
        self.pkt['I220']['RoC']['val'] = climbrate

    def setyawrate(self, yawrate):
        self.yawrate = yawrate
        
    def setspeed(self, speed):
        self.speed = speed

    def setheading(self, heading):
        self.heading = heading
        while self.heading > 360:
            self.heading -= 360.0
        while self.heading < 0:
            self.heading += 360.0

    def changealt(self, delta_alt):
        alt = self.pkt['I130']['Alt']['val']
        alt += delta_alt
        self.setalt(alt)

    def rate_of_turn(self, bank=45.0):
        '''return expected rate of turn in degrees/s for given speed in m/s and bank angle in degrees'''
        if abs(self.speed) < 2 or abs(bank) > 80:
            return 0
        ret = degrees(9.81*tan(radians(bank))/self.speed)
        return ret
        
    def update(self, deltat=1.0):
        self.move(self.heading, self.speed * deltat)
        climbrate = self.pkt['I220']['RoC']['val']
        self.changealt(climbrate * deltat)
        if self.desired_heading is None:
            self.setheading(self.heading + self.yawrate * deltat)
        else:
            heading_error = self.desired_heading - self.heading
            while heading_error > 180:
                heading_error -= 360.0
            while heading_error < -180:
                heading_error += 360.0            
            max_turn = self.rate_of_turn() * deltat
            if heading_error > 0:
                turn = min(max_turn, heading_error)
            else:
                turn = max(-max_turn, heading_error)
            self.setheading(self.heading + turn)
            if abs(heading_error) < 0.01:
                self.desired_heading = None

    def __str__(self):
        return str(self.pkt)

    def pickled(self):
        return b'PICKLED:' + pickle.dumps(self.pkt)

class Aircraft(DNFZ):
    '''an aircraft that flies in a circuit'''
    def __init__(self, speed=30.0, circuit_width=1000.0):
        DNFZ.__init__(self, 'Aircraft')
        self.setspeed(speed)
        self.circuit_width = circuit_width
        self.dist_flown = 0
        self.randalt()

    def update(self, deltat=1.0):
        '''fly a square circuit'''
        DNFZ.update(self, deltat)
        self.dist_flown += self.speed * deltat
        if self.dist_flown > self.circuit_width:
            self.desired_heading = self.heading + 90
            self.dist_flown = 0
        if self.getalt() < self.ground_height() or self.getalt() > self.ground_height() + 2000:
            self.randpos()
            self.randalt()

class BirdOfPrey(DNFZ):
    '''an bird that circles slowly climbing, then dives'''
    def __init__(self):
        DNFZ.__init__(self, 'BirdOfPrey')
        self.setspeed(16.0)
        self.radius = random.uniform(100,200)
        self.time_circling = 0
        self.dive_rate = -30
        self.climb_rate = 5
        self.drift_speed = random.uniform(5,10)
        self.max_alt = self.ground_height() + random.uniform(100, 400)
        self.drift_heading = self.heading
        circumference = math.pi * self.radius * 2
        circle_time = circumference / self.speed
        self.turn_rate = 360.0 / circle_time
        if random.uniform(0,1) < 0.5:
            self.turn_rate = -self.turn_rate

    def update(self, deltat=1.0):
        '''fly circles, then dive'''
        DNFZ.update(self, deltat)
        self.time_circling += deltat
        self.setheading(self.heading + self.turn_rate * deltat)
        self.move(self.drift_heading, self.drift_speed)
        if self.getalt() > self.max_alt or self.getalt() < self.ground_height():
            if self.getalt() > self.ground_height():
                self.setclimbrate(self.dive_rate)
            else:
                self.setclimbrate(self.climb_rate)
        if self.getalt() < self.ground_height():
            self.setalt(self.ground_height())
        if self.distance_from_home() > gen_settings.region_width:
            self.randpos()
            self.randalt()

class BirdMigrating(DNFZ):
    '''an bird that circles slowly climbing, then dives'''
    def __init__(self):
        DNFZ.__init__(self, 'BirdMigrating')
        self.setspeed(random.uniform(4,16))
        self.setyawrate(random.uniform(-0.2,0.2))
        self.randalt()

    def update(self, deltat=1.0):
        '''fly in long curves'''
        DNFZ.update(self, deltat)
        if (self.distance_from_home() > gen_settings.region_width or
            self.getalt() < self.ground_height() or
            self.getalt() > self.ground_height() + 1000):
            self.randpos()
            self.randalt()

class Weather(DNFZ):
    '''a weather system'''
    def __init__(self):
        DNFZ.__init__(self, 'Weather')
        self.setspeed(random.uniform(1,4))
        self.lifetime = random.uniform(300,600)
        self.setalt(0)
        self.setclimbrate(0)

    def update(self, deltat=1.0):
        '''straight lines, with short life'''
        DNFZ.update(self, deltat)
        self.lifetime -= deltat
        if self.lifetime <= 0:
            self.randpos()
            self.lifetime = random.uniform(300,600)

class GenobstaclesModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(GenobstaclesModule, self).__init__(mpstate, "genobstacles", "OBC 2018 obstacle generator")

        self.add_command('genobstacles', self.cmd_genobstacles, "obstacle generator",
                         ["<start|stop|restart|clearall|status>",
                          "set (GENSETTING)"])

        self.add_completion_function('(GENSETTING)',
                                     gen_settings.completion)
        self.sock = None
        self.aircraft = []
        self.last_t = 0
        self.menu_added_map = False
        self.pkt_queue = []
        self.have_home = False
        self.pending_start = True
        self.last_click = None
        if mp_util.has_wxpython:
            self.menu = MPMenuSubMenu('Obstacles',
                                    items=[MPMenuItem('Restart', 'Restart', '# genobstacles restart'),
                                           MPMenuItem('Stop', 'Stop', '# genobstacles stop'),
                                           MPMenuItem('Start','Start', '# genobstacles start'),
                                           MPMenuItem('Remove','Remove', '# genobstacles remove'),
                                           MPMenuItem('Drop Cloud','DropCloud', '# genobstacles dropcloud'),
                                           MPMenuItem('Drop Eagle','DropEagle', '# genobstacles dropeagle'),
                                           MPMenuItem('Drop Bird','DropBird', '# genobstacles dropbird'),
                                           MPMenuItem('Drop Plane','DropPlane', '# genobstacles dropplane'),
                                           MPMenuItem('ClearAll','ClearAll', '# genobstacles clearall')])

    def cmd_dropobject(self, obj):
        '''drop an object on the map'''
        latlon = self.mpstate.click_location
        if self.last_click is not None and self.last_click == latlon:
            return
        self.last_click = latlon
        if latlon is not None:
            obj.setpos(latlon[0], latlon[1])
            self.aircraft.append(obj)

    def status(self):
        ret = ""
        for aircraft in self.aircraft:
            ret += "%s %f %f\n" % (aircraft.DNFZ_type,
                                   aircraft.getlat(),
                                   aircraft.getlon(),)
        return ret

    def cmd_genobstacles(self, args):
        '''genobstacles command parser'''
        usage = "usage: genobstacles <start|stop|restart|clearall|status|set>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            gen_settings.command(args[1:])
        elif args[0] == "start":
            if self.have_home:
                self.start()
            else:
                self.pending_start = True
        elif args[0] == "stop":
            self.stop()
            self.pending_start = False
        elif args[0] == "restart":
            self.stop()
            self.start()
        elif args[0] == "status":
            print(self.status())
        elif args[0] == "remove":
            latlon = self.mpstate.click_location
            if self.last_click is not None and self.last_click == latlon:
                return
            self.last_click = latlon
            if latlon is not None:
                closest = None
                closest_distance = 1000
                for a in self.aircraft:
                    dist = a.distance_from(latlon[0], latlon[1])
                    if dist < closest_distance:
                        closest_distance = dist
                        closest = a
                if closest is not None:
                    self.aircraft.remove(closest)
                else:
                    print("No obstacle found at click point")
                    
        elif args[0] == "dropcloud":
            self.cmd_dropobject(Weather())
        elif args[0] == "dropeagle":
            self.cmd_dropobject(BirdOfPrey())
        elif args[0] == "dropbird":
            self.cmd_dropobject(BirdMigrating())
        elif args[0] == "dropplane":
            self.cmd_dropobject(Aircraft())
        elif args[0] == "clearall":
            self.clearall()
        else:
            print(usage)

    def start(self):
        '''start sending packets'''
        if self.sock is not None:
            self.sock.close()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.connect(('', gen_settings.port))

        global track_count
        self.aircraft = []
        track_count = 0
        self.last_t = 0

        # some fixed wing aircraft
        for i in range(gen_settings.num_aircraft):
            self.aircraft.append(Aircraft(random.uniform(10, 100), 2000.0))

        # some birds of prey
        for i in range(gen_settings.num_bird_prey):
            self.aircraft.append(BirdOfPrey())

        # some migrating birds
        for i in range(gen_settings.num_bird_migratory):
            self.aircraft.append(BirdMigrating())

        # some weather systems
        for i in range(gen_settings.num_weather):
            self.aircraft.append(Weather())
        print("Started on port %u" % gen_settings.port)

    def stop(self):
        '''stop listening for packets'''
        if self.sock is not None:
            self.sock.close()
            self.sock = None

    def clearall(self):
        '''remove all objects'''
        self.aircraft = []

    def idle_task(self):
        while len(self.pkt_queue) > 0:
            try:
                pkt = self.pkt_queue.pop(0)
                self.sock.send(pkt)
            except Exception as ex:
                return
            
    def mavlink_packet(self, m):
        '''trigger sends from ATTITUDE packets'''
        if not self.have_home and m.get_type() == 'GPS_RAW_INT' and m.fix_type >= 3:
            gen_settings.home_lat = m.lat * 1.0e-7
            gen_settings.home_lon = m.lon * 1.0e-7
            self.have_home = True
            if self.pending_start:
                self.start()
        if m.get_type() != 'ATTITUDE':
            return
        t = self.get_time()
        dt = t - self.last_t
        if dt < 0 or dt > 10:
            self.last_t = t
            return
        if dt > 10 or dt < 0.9:
            return
        self.last_t = t
        for a in self.aircraft:
            if not gen_settings.stop:
                a.update(1.0)
                self.pkt_queue.append(a.pickled())
                while len(self.pkt_queue) > len(self.aircraft)*2:
                    self.pkt_queue.pop(0)
                    
        if self.module('map') is not None and not self.menu_added_map:
            self.menu_added_map = True
            self.module('map').add_menu(self.menu)
        
        
def init(mpstate):
    '''initialise module'''
    return GenobstaclesModule(mpstate)



#!/usr/bin/env python
'''
antenna pointing module
Andrew Tridgell
June 2012
'''

import sys, os, time
from cuav.lib import cuav_util
from MAVProxy.modules.lib import mp_module

class AntennaModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(AntennaModule, self).__init__(mpstate, "antenna", "antenna pointing module")
        self.gcs_location = None
        self.last_bearing = 0
        self.last_announce = 0
        self.add_command('antenna', self.cmd_antenna, "antenna link control")

    def cmd_antenna(self, args):
        '''set gcs location'''
        if len(args) != 2:
            if self.gcs_location is None:
                print("GCS location not set")
            else:
                print("GCS location %s" % str(self.gcs_location))
            return
        self.gcs_location = (float(args[0]), float(args[1]))



    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if self.gcs_location is None and self.module('wp').wploader.count() > 0:
            home = self.module('wp').get_home()
            self.gcs_location = (home.x, home.y)
            print("Antenna home set")
        if self.gcs_location is None:
            return
        if m.get_type() == 'GPS_RAW' and self.gcs_location is not None:
            (gcs_lat, gcs_lon) = self.gcs_location
            bearing = cuav_util.gps_bearing(gcs_lat, gcs_lon, m.lat, m.lon)
        elif m.get_type() == 'GPS_RAW_INT' and self.gcs_location is not None:
            (gcs_lat, gcs_lon) = self.gcs_location
            bearing = cuav_util.gps_bearing(gcs_lat, gcs_lon, m.lat / 1.0e7, m.lon / 1.0e7)
        else:
            return
        self.console.set_status('Antenna', 'Antenna %.0f' % bearing, row=0)
        if abs(bearing - self.last_bearing) > 5 and (time.time() - self.last_announce) > 15:
            self.last_bearing = bearing
            self.last_announce = time.time()
            self.say("Antenna %u" % int(bearing + 0.5))

def init(mpstate):
    '''initialise module'''
    return AntennaModule(mpstate)



import time
from MAVProxy.modules.lib.wxhorizon_util import Attitude, VFR_HUD, Global_Position_INT, BatteryInfo, FlightState, WaypointInfo, FPS
from MAVProxy.modules.lib.wx_loader import wx
import math, time

import matplotlib
matplotlib.use('wxAgg')
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.pyplot import Polygon
import matplotlib.patheffects as PathEffects
from matplotlib import patches
import matplotlib as mpl

class HorizonFrame(wx.Frame):
    """ The main frame of the horizon indicator."""

    def __init__(self, state, title):
        self.state = state
        # Create Frame and Panel(s)
        wx.Frame.__init__(self, None, title=title)
        state.frame = self

        # Initialisation
        self.initData()
        self.initUI()
        self.startTime = time.time()
        self.nextTime = 0.0
        self.fps = 10.0

    def initData(self):
        # Initialise Attitude
        self.pitch = 0.0  # Degrees
        self.roll = 0.0   # Degrees
        self.yaw = 0.0    # Degrees
        
        # History Values
        self.oldRoll = 0.0 # Degrees
        
        # Initialise Rate Information
        self.airspeed = 0.0 # m/s
        self.relAlt = 0.0 # m relative to home position
        self.relAltTime = 0.0 # s The time that the relative altitude was recorded
        self.climbRate = 0.0 # m/s
        self.altHist = [] # Altitude History
        self.timeHist = [] # Time History
        self.altMax = 0.0 # Maximum altitude since startup
        
        # Initialise HUD Info
        self.heading = 0.0 # 0-360
        
        # Initialise Battery Info
        self.voltage = 0.0
        self.current = 0.0
        self.batRemain = 0.0

        # Initialise Mode and State
        self.mode = 'UNKNOWN'
        self.armed = ''
        self.safetySwitch = ''
        
        # Intialise Waypoint Information
        self.currentWP = 0
        self.finalWP = 0
        self.wpDist = 0
        self.nextWPTime = 0
        self.wpBearing = 0
    

    def initUI(self):
        # Create Event Timer and Bindings
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_timer, self.timer)
        self.timer.Start(100)
        self.Bind(wx.EVT_IDLE, self.on_idle)
        self.Bind(wx.EVT_CHAR_HOOK,self.on_KeyPress)

        # Create Panel
        self.panel = wx.Panel(self)
        self.vertSize = 0.09
        self.resized = False
        
        # Create Matplotlib Panel
        self.createPlotPanel()

        # Fix Axes - vertical is of length 2, horizontal keeps the same lengthscale
        self.rescaleX()
        self.calcFontScaling()        
        
        # Create Horizon Polygons
        self.createHorizonPolygons()
        
        # Center Pointer Marker
        self.thick = 0.015
        self.createCenterPointMarker()
        
        # Pitch Markers
        self.dist10deg = 0.2 # Graph distance per 10 deg
        self.createPitchMarkers()
        
        # Add Roll, Pitch, Yaw Text
        self.createRPYText()
        
        # Add Airspeed, Altitude, Climb Rate Text
        self.createAARText()
        
        # Create Heading Pointer
        self.createHeadingPointer()
        
        # Create North Pointer
        self.createNorthPointer()
        
        # Create Battery Bar
        self.batWidth = 0.1
        self.batHeight = 0.2
        self.rOffset = 0.35
        self.createBatteryBar()
        
        # Create Mode & State Text
        self.createStateText()
        
        # Create Waypoint Text
        self.createWPText()
        
        # Create Waypoint Pointer
        self.createWPPointer()
        
        # Create Altitude History Plot
        self.createAltHistoryPlot()
        
        # Show Frame
        self.Show(True)
        self.pending = []
    
    def createPlotPanel(self):
        '''Creates the figure and axes for the plotting panel.'''
        self.figure = Figure()
        self.axes = self.figure.add_subplot(111)
        self.canvas = FigureCanvas(self,-1,self.figure)
        self.canvas.SetSize(wx.Size(300,300))
        self.axes.axis('off')
        self.figure.subplots_adjust(left=0,right=1,top=1,bottom=0)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.canvas,1,wx.EXPAND,wx.ALL)
        self.SetSizerAndFit(self.sizer)
        self.Fit()
        
    def rescaleX(self):
        '''Rescales the horizontal axes to make the lengthscales equal.'''
        self.ratio = self.figure.get_size_inches()[0]/float(self.figure.get_size_inches()[1])
        self.axes.set_xlim(-self.ratio,self.ratio)
        self.axes.set_ylim(-1,1)
        
    def calcFontScaling(self):
        '''Calculates the current font size and left position for the current window.'''
        self.ypx = self.figure.get_size_inches()[1]*self.figure.dpi
        self.xpx = self.figure.get_size_inches()[0]*self.figure.dpi
        self.fontSize = self.vertSize*(self.ypx/2.0)
        self.leftPos = self.axes.get_xlim()[0]
        self.rightPos = self.axes.get_xlim()[1]
    
    def checkReszie(self):
        '''Checks if the window was resized.'''
        if not self.resized:
            oldypx = self.ypx
            oldxpx = self.xpx
            self.ypx = self.figure.get_size_inches()[1]*self.figure.dpi
            self.xpx = self.figure.get_size_inches()[0]*self.figure.dpi
            if (oldypx != self.ypx) or (oldxpx != self.xpx):
                self.resized = True
            else:
                self.resized = False
    
    def createHeadingPointer(self):
        '''Creates the pointer for the current heading.'''
        self.headingTri = patches.RegularPolygon((0.0,0.80),3,0.05,color='k',zorder=4)
        self.axes.add_patch(self.headingTri)
        self.headingText = self.axes.text(0.0,0.675,'0',color='k',size=self.fontSize,horizontalalignment='center',verticalalignment='center',zorder=4)
    
    def adjustHeadingPointer(self):
        '''Adjust the value of the heading pointer.'''
        self.headingText.set_text(str(self.heading))
        self.headingText.set_size(self.fontSize) 
    
    def createNorthPointer(self):
        '''Creates the north pointer relative to current heading.'''
        self.headingNorthTri = patches.RegularPolygon((0.0,0.80),3,0.05,color='k',zorder=4)
        self.axes.add_patch(self.headingNorthTri)
        self.headingNorthText = self.axes.text(0.0,0.675,'N',color='k',size=self.fontSize,horizontalalignment='center',verticalalignment='center',zorder=4)    

    def adjustNorthPointer(self):
        '''Adjust the position and orientation of
        the north pointer.'''
        self.headingNorthText.set_size(self.fontSize) 
        headingRotate = mpl.transforms.Affine2D().rotate_deg_around(0.0,0.0,self.heading)+self.axes.transData
        self.headingNorthText.set_transform(headingRotate)
        if (self.heading > 90) and (self.heading < 270):
            headRot = self.heading-180
        else:
            headRot = self.heading
        self.headingNorthText.set_rotation(headRot)
        self.headingNorthTri.set_transform(headingRotate)
        # Adjust if overlapping with heading pointer
        if (self.heading <= 10.0) or (self.heading >= 350.0):
            self.headingNorthText.set_text('')
        else:
            self.headingNorthText.set_text('N')
            
    def toggleWidgets(self,widgets):
        '''Hides/shows the given widgets.'''
        for wig in widgets:
            if wig.get_visible():
                wig.set_visible(False)
            else:
                wig.set_visible(True)
    
    def createRPYText(self):
        '''Creates the text for roll, pitch and yaw.'''
        self.rollText = self.axes.text(self.leftPos+(self.vertSize/10.0),-0.97+(2*self.vertSize)-(self.vertSize/10.0),'Roll:   %.2f' % self.roll,color='w',size=self.fontSize)
        self.pitchText = self.axes.text(self.leftPos+(self.vertSize/10.0),-0.97+self.vertSize-(0.5*self.vertSize/10.0),'Pitch: %.2f' % self.pitch,color='w',size=self.fontSize)
        self.yawText = self.axes.text(self.leftPos+(self.vertSize/10.0),-0.97,'Yaw:   %.2f' % self.yaw,color='w',size=self.fontSize)
        self.rollText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        self.pitchText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        self.yawText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        
    def updateRPYLocations(self):
        '''Update the locations of roll, pitch, yaw text.'''
        # Locations
        self.rollText.set_position((self.leftPos+(self.vertSize/10.0),-0.97+(2*self.vertSize)-(self.vertSize/10.0)))
        self.pitchText.set_position((self.leftPos+(self.vertSize/10.0),-0.97+self.vertSize-(0.5*self.vertSize/10.0)))
        self.yawText.set_position((self.leftPos+(self.vertSize/10.0),-0.97))
        # Font Size
        self.rollText.set_size(self.fontSize)
        self.pitchText.set_size(self.fontSize)
        self.yawText.set_size(self.fontSize)
        
    def updateRPYText(self):
        'Updates the displayed Roll, Pitch, Yaw Text'
        self.rollText.set_text('Roll:   %.2f' % self.roll)
        self.pitchText.set_text('Pitch: %.2f' % self.pitch)
        self.yawText.set_text('Yaw:   %.2f' % self.yaw)
        
    def createCenterPointMarker(self):
        '''Creates the center pointer in the middle of the screen.'''
        self.axes.add_patch(patches.Rectangle((-0.75,-self.thick),0.5,2.0*self.thick,facecolor='orange',zorder=3))
        self.axes.add_patch(patches.Rectangle((0.25,-self.thick),0.5,2.0*self.thick,facecolor='orange',zorder=3))
        self.axes.add_patch(patches.Circle((0,0),radius=self.thick,facecolor='orange',edgecolor='none',zorder=3))
        
    def createHorizonPolygons(self):
        '''Creates the two polygons to show the sky and ground.'''
        # Sky Polygon
        vertsTop = [[-1,0],[-1,1],[1,1],[1,0],[-1,0]]
        self.topPolygon = Polygon(vertsTop,facecolor='dodgerblue',edgecolor='none')
        self.axes.add_patch(self.topPolygon)
        # Ground Polygon
        vertsBot = [[-1,0],[-1,-1],[1,-1],[1,0],[-1,0]]
        self.botPolygon = Polygon(vertsBot,facecolor='brown',edgecolor='none')
        self.axes.add_patch(self.botPolygon)
        
    def calcHorizonPoints(self):
        '''Updates the verticies of the patches for the ground and sky.'''
        ydiff = math.tan(math.radians(-self.roll))*float(self.ratio)
        pitchdiff = self.dist10deg*(self.pitch/10.0)
        # Draw Polygons
        vertsTop = [(-self.ratio,ydiff-pitchdiff),(-self.ratio,1),(self.ratio,1),(self.ratio,-ydiff-pitchdiff),(-self.ratio,ydiff-pitchdiff)]
        vertsBot = [(-self.ratio,ydiff-pitchdiff),(-self.ratio,-1),(self.ratio,-1),(self.ratio,-ydiff-pitchdiff),(-self.ratio,ydiff-pitchdiff)]
        if (self.roll > 90) or (self.roll < -90):
            vertsTop = [(-self.ratio,ydiff-pitchdiff),(-self.ratio,-1),(self.ratio,-1),(self.ratio,-ydiff-pitchdiff),(-self.ratio,ydiff-pitchdiff)]
            vertsBot = [(-self.ratio,ydiff-pitchdiff),(-self.ratio,1),(self.ratio,1),(self.ratio,-ydiff-pitchdiff),(-self.ratio,ydiff-pitchdiff)]
        self.topPolygon.set_xy(vertsTop)
        self.botPolygon.set_xy(vertsBot)  
    
    def createPitchMarkers(self):
        '''Creates the rectangle patches for the pitch indicators.'''
        self.pitchPatches = []
        # Major Lines (multiple of 10 deg)
        for i in [-9,-8,-7,-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6,7,8,9]:
            width = self.calcPitchMarkerWidth(i)
            currPatch = patches.Rectangle((-width/2.0,self.dist10deg*i-(self.thick/2.0)),width,self.thick,facecolor='w',edgecolor='none')
            self.axes.add_patch(currPatch)
            self.pitchPatches.append(currPatch)
        # Add Label for +-30 deg
        self.vertSize = 0.09
        self.pitchLabelsLeft = []
        self.pitchLabelsRight = []
        i=0
        for j in [-90,-60,-30,30,60,90]:
            self.pitchLabelsLeft.append(self.axes.text(-0.55,(j/10.0)*self.dist10deg,str(j),color='w',size=self.fontSize,horizontalalignment='center',verticalalignment='center'))
            self.pitchLabelsLeft[i].set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
            self.pitchLabelsRight.append(self.axes.text(0.55,(j/10.0)*self.dist10deg,str(j),color='w',size=self.fontSize,horizontalalignment='center',verticalalignment='center'))
            self.pitchLabelsRight[i].set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
            i += 1
        
    def calcPitchMarkerWidth(self,i):
        '''Calculates the width of a pitch marker.'''
        if (i % 3) == 0:
            if i == 0:
                width = 1.5
            else:
                width = 0.9
        else:
            width = 0.6
            
        return width
        
    def adjustPitchmarkers(self):
        '''Adjusts the location and orientation of pitch markers.'''
        pitchdiff = self.dist10deg*(self.pitch/10.0)
        rollRotate = mpl.transforms.Affine2D().rotate_deg_around(0.0,-pitchdiff,self.roll)+self.axes.transData
        j=0
        for i in [-9,-8,-7,-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6,7,8,9]:
            width = self.calcPitchMarkerWidth(i)
            self.pitchPatches[j].set_xy((-width/2.0,self.dist10deg*i-(self.thick/2.0)-pitchdiff))
            self.pitchPatches[j].set_transform(rollRotate)
            j+=1
        # Adjust Text Size and rotation
        i=0
        for j in [-9,-6,-3,3,6,9]:
                self.pitchLabelsLeft[i].set_y(j*self.dist10deg-pitchdiff)
                self.pitchLabelsRight[i].set_y(j*self.dist10deg-pitchdiff)
                self.pitchLabelsLeft[i].set_size(self.fontSize)
                self.pitchLabelsRight[i].set_size(self.fontSize)
                self.pitchLabelsLeft[i].set_rotation(self.roll)
                self.pitchLabelsRight[i].set_rotation(self.roll)
                self.pitchLabelsLeft[i].set_transform(rollRotate)
                self.pitchLabelsRight[i].set_transform(rollRotate)
                i += 1 
                
    def createAARText(self):
        '''Creates the text for airspeed, altitude and climb rate.'''
        self.airspeedText = self.axes.text(self.rightPos-(self.vertSize/10.0),-0.97+(2*self.vertSize)-(self.vertSize/10.0),'AS:   %.1f m/s' % self.airspeed,color='w',size=self.fontSize,ha='right')
        self.altitudeText = self.axes.text(self.rightPos-(self.vertSize/10.0),-0.97+self.vertSize-(0.5*self.vertSize/10.0),'ALT: %.1f m   ' % self.relAlt,color='w',size=self.fontSize,ha='right')
        self.climbRateText = self.axes.text(self.rightPos-(self.vertSize/10.0),-0.97,'CR:   %.1f m/s' % self.climbRate,color='w',size=self.fontSize,ha='right')
        self.airspeedText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        self.altitudeText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        self.climbRateText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        
    def updateAARLocations(self):
        '''Update the locations of airspeed, altitude and Climb rate.'''
        # Locations
        self.airspeedText.set_position((self.rightPos-(self.vertSize/10.0),-0.97+(2*self.vertSize)-(self.vertSize/10.0)))
        self.altitudeText.set_position((self.rightPos-(self.vertSize/10.0),-0.97+self.vertSize-(0.5*self.vertSize/10.0)))
        self.climbRateText.set_position((self.rightPos-(self.vertSize/10.0),-0.97))
        # Font Size
        self.airspeedText.set_size(self.fontSize)
        self.altitudeText.set_size(self.fontSize)
        self.climbRateText.set_size(self.fontSize)
        
    def updateAARText(self):
        'Updates the displayed airspeed, altitude, climb rate Text'
        self.airspeedText.set_text('AR:   %.1f m/s' % self.airspeed)
        self.altitudeText.set_text('ALT: %.1f m   ' % self.relAlt)
        self.climbRateText.set_text('CR:   %.1f m/s' % self.climbRate)
    
    def createBatteryBar(self):
        '''Creates the bar to display current battery percentage.'''
        self.batOutRec = patches.Rectangle((self.rightPos-(1.3+self.rOffset)*self.batWidth,1.0-(0.1+1.0+(2*0.075))*self.batHeight),self.batWidth*1.3,self.batHeight*1.15,facecolor='darkgrey',edgecolor='none')
        self.batInRec = patches.Rectangle((self.rightPos-(self.rOffset+1+0.15)*self.batWidth,1.0-(0.1+1+0.075)*self.batHeight),self.batWidth,self.batHeight,facecolor='lawngreen',edgecolor='none')
        self.batPerText = self.axes.text(self.rightPos - (self.rOffset+0.65)*self.batWidth,1-(0.1+1+(0.075+0.15))*self.batHeight,'%.f' % self.batRemain,color='w',size=self.fontSize,ha='center',va='top')
        self.batPerText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        self.voltsText = self.axes.text(self.rightPos-(self.rOffset+1.3+0.2)*self.batWidth,1-(0.1+0.05+0.075)*self.batHeight,'%.1f V' % self.voltage,color='w',size=self.fontSize,ha='right',va='top')
        self.ampsText = self.axes.text(self.rightPos-(self.rOffset+1.3+0.2)*self.batWidth,1-self.vertSize-(0.1+0.05+0.1+0.075)*self.batHeight,'%.1f A' % self.current,color='w',size=self.fontSize,ha='right',va='top')
        self.voltsText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        self.ampsText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')])
        
        self.axes.add_patch(self.batOutRec)
        self.axes.add_patch(self.batInRec)
        
    def updateBatteryBar(self):
        '''Updates the position and values of the battery bar.'''
        # Bar
        self.batOutRec.set_xy((self.rightPos-(1.3+self.rOffset)*self.batWidth,1.0-(0.1+1.0+(2*0.075))*self.batHeight))
        self.batInRec.set_xy((self.rightPos-(self.rOffset+1+0.15)*self.batWidth,1.0-(0.1+1+0.075)*self.batHeight))
        self.batPerText.set_position((self.rightPos - (self.rOffset+0.65)*self.batWidth,1-(0.1+1+(0.075+0.15))*self.batHeight))
        self.batPerText.set_fontsize(self.fontSize)
        self.voltsText.set_text('%.1f V' % self.voltage)
        self.ampsText.set_text('%.1f A' % self.current)
        self.voltsText.set_position((self.rightPos-(self.rOffset+1.3+0.2)*self.batWidth,1-(0.1+0.05)*self.batHeight))
        self.ampsText.set_position((self.rightPos-(self.rOffset+1.3+0.2)*self.batWidth,1-self.vertSize-(0.1+0.05+0.1)*self.batHeight))
        self.voltsText.set_fontsize(self.fontSize)
        self.ampsText.set_fontsize(self.fontSize)
        if self.batRemain >= 0:
            self.batPerText.set_text(int(self.batRemain))
            self.batInRec.set_height(self.batRemain*self.batHeight/100.0)
            if self.batRemain/100.0 > 0.5:
                self.batInRec.set_facecolor('lawngreen')
            elif self.batRemain/100.0 <= 0.5 and self.batRemain/100.0 > 0.2:
                self.batInRec.set_facecolor('yellow')
            elif self.batRemain/100.0 <= 0.2 and self.batRemain >= 0.0:
                self.batInRec.set_facecolor('r')
        elif self.batRemain == -1:
            self.batInRec.set_height(self.batHeight)
            self.batInRec.set_facecolor('k')
        
    def createStateText(self):
        '''Creates the mode and arm state text.'''
        self.modeText = self.axes.text(self.leftPos+(self.vertSize/10.0),0.97,'UNKNOWN',color='grey',size=1.5*self.fontSize,ha='left',va='top')
        self.modeText.set_path_effects([PathEffects.withStroke(linewidth=self.fontSize/10.0,foreground='black')])
        
    def updateStateText(self):
        '''Updates the mode and colours red or green depending on arm state.'''
        self.modeText.set_position((self.leftPos+(self.vertSize/10.0),0.97))
        self.modeText.set_text(self.mode)
        self.modeText.set_size(1.5*self.fontSize)
        if self.armed:
            self.modeText.set_color('red')
            self.modeText.set_path_effects([PathEffects.withStroke(linewidth=self.fontSize/10.0,foreground='yellow')])
        elif (self.armed == False):
            self.modeText.set_color('lightgreen')
            self.modeText.set_bbox(None)
            self.modeText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='black')])
        else:
            # Fall back if unknown
            self.modeText.set_color('grey')
            self.modeText.set_bbox(None)
            self.modeText.set_path_effects([PathEffects.withStroke(linewidth=self.fontSize/10.0,foreground='black')])
        
    def createWPText(self):
        '''Creates the text for the current and final waypoint,
        and the distance to the new waypoint.'''
        self.wpText = self.axes.text(self.leftPos+(1.5*self.vertSize/10.0),0.97-(1.5*self.vertSize)+(0.5*self.vertSize/10.0),'0/0\n(0 m, 0 s)',color='w',size=self.fontSize,ha='left',va='top')
        self.wpText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='black')])
        
    def updateWPText(self):
        '''Updates the current waypoint and distance to it.''' 
        self.wpText.set_position((self.leftPos+(1.5*self.vertSize/10.0),0.97-(1.5*self.vertSize)+(0.5*self.vertSize/10.0)))
        self.wpText.set_size(self.fontSize)
        if type(self.nextWPTime) is str:
            self.wpText.set_text('%.f/%.f\n(%.f m, ~ s)' % (self.currentWP,self.finalWP,self.wpDist))
        else:
            self.wpText.set_text('%.f/%.f\n(%.f m, %.f s)' % (self.currentWP,self.finalWP,self.wpDist,self.nextWPTime))
    
    def createWPPointer(self):
        '''Creates the waypoint pointer relative to current heading.'''
        self.headingWPTri = patches.RegularPolygon((0.0,0.55),3,0.05,facecolor='lime',zorder=4,ec='k')
        self.axes.add_patch(self.headingWPTri)
        self.headingWPText = self.axes.text(0.0,0.45,'1',color='lime',size=self.fontSize,horizontalalignment='center',verticalalignment='center',zorder=4)   
        self.headingWPText.set_path_effects([PathEffects.withStroke(linewidth=1,foreground='k')]) 

    def adjustWPPointer(self):
        '''Adjust the position and orientation of
        the waypoint pointer.'''
        self.headingWPText.set_size(self.fontSize) 
        headingRotate = mpl.transforms.Affine2D().rotate_deg_around(0.0,0.0,-self.wpBearing+self.heading)+self.axes.transData
        self.headingWPText.set_transform(headingRotate)
        angle = self.wpBearing - self.heading
        if angle < 0:
            angle += 360
        if (angle > 90) and (angle < 270):
            headRot = angle-180
        else:
            headRot = angle
        self.headingWPText.set_rotation(-headRot)
        self.headingWPTri.set_transform(headingRotate)
        self.headingWPText.set_text('%.f' % (angle))
    
    def createAltHistoryPlot(self):
        '''Creates the altitude history plot.'''
        self.altHistRect = patches.Rectangle((self.leftPos+(self.vertSize/10.0),-0.25),0.5,0.5,facecolor='grey',edgecolor='none',alpha=0.4,zorder=4)
        self.axes.add_patch(self.altHistRect)
        self.altPlot, = self.axes.plot([self.leftPos+(self.vertSize/10.0),self.leftPos+(self.vertSize/10.0)+0.5],[0.0,0.0],color='k',marker=None,zorder=4)
        self.altMarker, = self.axes.plot(self.leftPos+(self.vertSize/10.0)+0.5,0.0,marker='o',color='k',zorder=4)
        self.altText2 = self.axes.text(self.leftPos+(4*self.vertSize/10.0)+0.5,0.0,'%.f m' % self.relAlt,color='k',size=self.fontSize,ha='left',va='center',zorder=4)
    
    def updateAltHistory(self):
        '''Updates the altitude history plot.'''
        self.altHist.append(self.relAlt)
        self.timeHist.append(self.relAltTime)
        
        # Delete entries older than x seconds
        histLim = 10
        currentTime = time.time()
        point = 0
        for i in range(0,len(self.timeHist)):
            if (self.timeHist[i] > (currentTime - 10.0)):
                break
        # Remove old entries
        self.altHist = self.altHist[i:]
        self.timeHist = self.timeHist[i:]
        
        # Transform Data
        x = []
        y = []
        tmin = min(self.timeHist)
        tmax = max(self.timeHist)
        x1 = self.leftPos+(self.vertSize/10.0)
        y1 = -0.25
        altMin = 0
        altMax = max(self.altHist)
        # Keep alt max for whole mission
        if altMax > self.altMax:
            self.altMax = altMax
        else:
            altMax = self.altMax
        if tmax != tmin:
            mx = 0.5/(tmax-tmin)
        else:
            mx = 0.0
        if altMax != altMin:
            my = 0.5/(altMax-altMin)
        else:
            my = 0.0
        for t in self.timeHist:
            x.append(mx*(t-tmin)+x1)
        for alt in self.altHist:
            val = my*(alt-altMin)+y1
            # Crop extreme noise
            if val < -0.25:
                val = -0.25
            elif val > 0.25:
                val = 0.25
            y.append(val)
        # Display Plot
        self.altHistRect.set_x(self.leftPos+(self.vertSize/10.0))
        self.altPlot.set_data(x,y)
        self.altMarker.set_data(self.leftPos+(self.vertSize/10.0)+0.5,val)
        self.altText2.set_position((self.leftPos+(4*self.vertSize/10.0)+0.5,val))
        self.altText2.set_size(self.fontSize)
        self.altText2.set_text('%.f m' % self.relAlt)
        
    # =============== Event Bindings =============== #    
    def on_idle(self, event):
        '''To adjust text and positions on rescaling the window when resized.'''
        # Check for resize
        self.checkReszie()
        
        if self.resized:
            # Fix Window Scales 
            self.rescaleX()
            self.calcFontScaling()
            
            # Recalculate Horizon Polygons
            self.calcHorizonPoints()
            
            # Update Roll, Pitch, Yaw Text Locations
            self.updateRPYLocations()
            
            # Update Airpseed, Altitude, Climb Rate Locations
            self.updateAARLocations()
            
            # Update Pitch Markers
            self.adjustPitchmarkers()
            
            # Update Heading and North Pointer
            self.adjustHeadingPointer()
            self.adjustNorthPointer()
            
            # Update Battery Bar
            self.updateBatteryBar()
            
            # Update Mode and State
            self.updateStateText()
            
            # Update Waypoint Text
            self.updateWPText()
            
            # Adjust Waypoint Pointer
            self.adjustWPPointer()
            
            # Update History Plot
            self.updateAltHistory()
            
            # Update Matplotlib Plot
            self.canvas.draw()
            self.canvas.Refresh()
            
            self.resized = False
        
        time.sleep(0.05)
 
    def on_timer(self, event):
        '''Main Loop.'''
        state = self.state
        self.loopStartTime = time.time()
        if state.close_event.wait(0.001):
            self.timer.Stop()
            self.Destroy()
            return
        
        # Check for resizing
        self.checkReszie()
        if self.resized:
            self.on_idle(0)
        
        # Get attitude information
        while state.child_pipe_recv.poll():           
            objList = state.child_pipe_recv.recv()
            for obj in objList:
                self.calcFontScaling()
                if isinstance(obj,Attitude):
                    self.oldRoll = self.roll
                    self.pitch = obj.pitch*180/math.pi
                    self.roll = obj.roll*180/math.pi
                    self.yaw = obj.yaw*180/math.pi
                    
                    # Update Roll, Pitch, Yaw Text Text
                    self.updateRPYText()
                    
                    # Recalculate Horizon Polygons
                    self.calcHorizonPoints()
                    
                    # Update Pitch Markers
                    self.adjustPitchmarkers()
                
                elif isinstance(obj,VFR_HUD):
                    self.heading = obj.heading
                    self.airspeed = obj.airspeed
                    self.climbRate = obj.climbRate
                    
                    # Update Airpseed, Altitude, Climb Rate Locations
                    self.updateAARText()
                    
                    # Update Heading North Pointer
                    self.adjustHeadingPointer()
                    self.adjustNorthPointer()
                
                elif isinstance(obj,Global_Position_INT):
                    self.relAlt = obj.relAlt
                    self.relAltTime = obj.curTime
                    
                    # Update Airpseed, Altitude, Climb Rate Locations
                    self.updateAARText()
                    
                    # Update Altitude History
                    self.updateAltHistory()
                    
                elif isinstance(obj,BatteryInfo):
                    self.voltage = obj.voltage
                    self.current = obj.current
                    self.batRemain = obj.batRemain
                    
                    # Update Battery Bar
                    self.updateBatteryBar()
                    
                elif isinstance(obj,FlightState):
                    self.mode = obj.mode
                    self.armed = obj.armState
                    
                    # Update Mode and Arm State Text
                    self.updateStateText()
                    
                elif isinstance(obj,WaypointInfo):
                    self.currentWP = obj.current
                    self.finalWP = obj.final
                    self.wpDist = obj.currentDist
                    self.nextWPTime = obj.nextWPTime
                    if obj.wpBearing < 0.0:
                        self.wpBearing = obj.wpBearing + 360
                    else:
                        self.wpBearing = obj.wpBearing
                    
                    # Update waypoint text
                    self.updateWPText()
                    
                    # Adjust Waypoint Pointer
                    self.adjustWPPointer()
                    
                elif isinstance(obj, FPS):
                    # Update fps target
                    self.fps = obj.fps
                
                
        # Quit Drawing if too early
        if (time.time() > self.nextTime):                     
            # Update Matplotlib Plot
            self.canvas.draw()
            self.canvas.Refresh()                 
                 
            self.Refresh()
            self.Update()
            
            # Calculate next frame time
            if (self.fps > 0):
                fpsTime = 1/self.fps
                self.nextTime = fpsTime + self.loopStartTime
            else:
                self.nextTime = time.time()
                
    def on_KeyPress(self,event):
        '''To adjust the distance between pitch markers.'''
        if event.GetKeyCode() == wx.WXK_UP:
            self.dist10deg += 0.1
            print('Dist per 10 deg: %.1f' % self.dist10deg)
        elif event.GetKeyCode() == wx.WXK_DOWN:
            self.dist10deg -= 0.1
            if self.dist10deg <= 0:
                self.dist10deg = 0.1
            print('Dist per 10 deg: %.1f' % self.dist10deg)
        # Toggle Widgets
        elif event.GetKeyCode() == 49: # 1
            widgets = [self.modeText,self.wpText]
            self.toggleWidgets(widgets)  
        elif event.GetKeyCode() == 50: # 2
            widgets = [self.batOutRec,self.batInRec,self.voltsText,self.ampsText,self.batPerText]
            self.toggleWidgets(widgets)     
        elif event.GetKeyCode() == 51: # 3
            widgets = [self.rollText,self.pitchText,self.yawText]
            self.toggleWidgets(widgets)                 
        elif event.GetKeyCode() == 52: # 4
            widgets = [self.airspeedText,self.altitudeText,self.climbRateText]
            self.toggleWidgets(widgets)
        elif event.GetKeyCode() == 53: # 5
            widgets = [self.altHistRect,self.altPlot,self.altMarker,self.altText2]
            self.toggleWidgets(widgets)
        elif event.GetKeyCode() == 54: # 6
            widgets = [self.headingTri,self.headingText,self.headingNorthTri,self.headingNorthText,self.headingWPTri,self.headingWPText]
            self.toggleWidgets(widgets)
            
        # Update Matplotlib Plot
        self.canvas.draw()
        self.canvas.Refresh()                       
        
        self.Refresh()
        self.Update()



#!/usr/bin/env python
"""
 NTRIP client module
 based on client from http://github.com/jcmb/NTRIP
"""

import socket
import sys
import datetime
import base64
import time
import errno
import select
from MAVProxy.modules.lib import rtcm3
import ssl
from optparse import OptionParser

version = 0.1
useragent = "NTRIP MAVProxy/%.1f" % version


class NtripError(Exception):
    def __init__(self, message, inner_exception=None):
        self.message = message
        self.inner_exception = inner_exception
        self.exception_info = sys.exc_info()

    def __str__(self):
        return self.message


class NtripClient(object):
    def __init__(self,
                 user="",
                 port=2101,
                 caster="",
                 mountpoint="",
                 host=False,
                 lat=46,
                 lon=122,
                 height=1212,
                 ssl=False,
                 V2=False,
                 ):
        if sys.version_info.major >= 3:
            user = bytearray(user, 'ascii')
        self.user = base64.b64encode(user)
        self.port = port
        self.caster = caster
        self.caster_ip = None
        self.mountpoint = mountpoint
        if not self.mountpoint.startswith("/"):
            self.mountpoint = "/" + self.mountpoint
        self.setPosition(lat, lon)
        self.height = height
        self.ssl = ssl
        self.host = host
        self.V2 = V2
        self.socket = None
        self.socket_pending = None
        self.found_header = False
        self.sent_header = False
        # RTCM3 parser
        self.rtcm3 = rtcm3.RTCM3()
        self.last_id = None
        self.dt_last_gga_sent = 0
        self.last_connect_attempt = time.time()
        if self.port == 443:
            # force SSL on port 443
            self.ssl = True

    def setPosition(self, lat, lon):
        self.flagN = "N"
        self.flagE = "E"
        if lon > 180:
            lon = (lon-360)*-1
            self.flagE = "W"
        elif lon < 0 and lon >= -180:
            lon = lon*-1
            self.flagE = "W"
        elif lon < -180:
            lon = lon+360
            self.flagE = "E"
        else:
            self.lon = lon
        if lat < 0:
            lat = lat*-1
            self.flagN = "S"
        self.lonDeg = int(lon)
        self.latDeg = int(lat)
        self.lonMin = (lon-self.lonDeg)*60
        self.latMin = (lat-self.latDeg)*60

    def getMountPointString(self):
        userstr = self.user
        if sys.version_info.major >= 3:
            userstr = str(userstr, 'ascii')
        mountPointString = "GET %s HTTP/1.0\r\nUser-Agent: %s\r\nAuthorization: Basic %s\r\n" % (
            self.mountpoint, useragent, userstr)
        if self.host or self.V2:
            hostString = "Host: %s:%i\r\n" % (self.caster, self.port)
            mountPointString += hostString
        if self.V2:
            mountPointString += "Ntrip-Version: Ntrip/2.0\r\n"
        mountPointString += "\r\n"
        return mountPointString

    def getGGAString(self):
        now = datetime.datetime.utcnow()
        ggaString = "GPGGA,%02d%02d%04.2f,%02d%011.8f,%1s,%03d%011.8f,%1s,1,05,0.19,+00400,M,%5.3f,M,," % (
                     now.hour, now.minute, now.second, self.latDeg, self.latMin, self.flagN,
                     self.lonDeg, self.lonMin, self.flagE, self.height)
        checksum = self.calculateCheckSum(ggaString)
        return "$%s*%s\r\n" % (ggaString, checksum)

    def calculateCheckSum(self, stringToCheck):
        xsum_calc = 0
        for char in stringToCheck:
            xsum_calc = xsum_calc ^ ord(char)
        return "%02X" % xsum_calc

    def get_ID(self):
        '''get ID of last packet'''
        return self.last_id

    def read(self):
        if self.socket is None:
            if self.socket_pending is None:
                now = time.time()
                # rate limit connection attempts
                if now - self.last_connect_attempt < 1.0:
                    return None
                self.last_connect_attempt = now
            self.connect()
            return None

        if not self.found_header:
            if not self.sent_header:
                self.sent_header = True
                time.sleep(0.1)
                mps = self.getMountPointString()
                if sys.version_info.major >= 3:
                    mps = bytearray(mps, 'ascii')
                try:
                    self.socket.sendall(mps)
                except Exception:
                    self.socket = None
                    return None
            try:
                casterResponse = self.socket.recv(4096)
            except ssl.SSLWantReadError:
                    return None
            except IOError as e:
                if e.errno == errno.EWOULDBLOCK:
                    return None
                self.socket = None
                casterResponse = ''
            if sys.version_info.major >= 3:
                # Ignore non ascii characters in HTTP response
                casterResponse = str(casterResponse, 'ascii', 'ignore')
            header_lines = casterResponse.split("\r\n")
            for line in header_lines:
                if line == "":
                    self.found_header = True
                if line.find("SOURCETABLE") != -1:
                    raise NtripError("Mount point does not exist")
                elif line.find("401 Unauthorized") != -1:
                    raise NtripError("Unauthorized request")
                elif line.find("404 Not Found") != -1:
                    raise NtripError("Mount Point does not exist")
                elif line.find(" 200 OK") != -1:
                    # Request was valid
                    self.send_gga()
            return None
        # normal data read
        while True:
            try:
                data = self.socket.recv(1)
            except ssl.SSLWantReadError:
                    return None
            except IOError as e:
                if e.errno == errno.EWOULDBLOCK:
                    return None
                self.socket.close()
                self.socket = None
                return None
            except Exception:
                self.socket.close()
                self.socket = None
                return None
            if len(data) == 0:
                self.socket.close()
                self.socket = None
                return None
            if self.rtcm3.read(data):
                self.last_id = self.rtcm3.get_packet_ID()
                return self.rtcm3.get_packet()

    def connect(self):
        '''connect to NTRIP server'''
        self.sent_header = False
        self.found_header = False
        # use a non-blocking connect to prevent blocking of mavproxy main thread
        if self.socket_pending is not None:
            read_ok, write_ok, err = select.select([], [self.socket_pending], [], 0)
            if self.socket_pending in write_ok:
                err = self.socket_pending.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
                if err == 0:
                    self.socket = self.socket_pending
                    self.socket_pending = None
                    self.rtcm3.reset()
                    return True
                else:
                    self.socket_pending = None
            return
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(0)
        if self.ssl:
            sock = ssl.wrap_socket(sock)
        try:
            if self.caster_ip is None:
                try:
                    self.caster_ip = socket.gethostbyname(self.caster)
                except Exception:
                    return False
            if self.caster_ip is None:
                return False
            error_indicator = sock.connect_ex((self.caster_ip, self.port))
            self.socket_pending = sock
            return False
        except Exception as ex:
            return False
        if error_indicator == 0:
            sock.setblocking(0)
            self.socket = sock
            self.rtcm3.reset()
            return True
        return False

    def readLoop(self):
        while True:
            data = self.read()
            if data is None:
                continue
            print("got: ", len(data))

    def send_gga(self):
        gga = self.getGGAString()
        if sys.version_info.major >= 3:
            gga = bytearray(gga, "ascii")
        try:
            self.socket.sendall(gga)
            self.dt_last_gga_sent = time.time()
        except Exception:
            self.socket = None

if __name__ == '__main__':
    usage = "NtripClient.py [options] [caster] [port] mountpoint"
    parser = OptionParser(version=version, usage=usage)
    parser.add_option("-u", "--user", type="string", dest="user", default="IBS", help="The Ntripcaster username.  Default: %default")
    parser.add_option("-p", "--password", type="string", dest="password", default="IBS", help="The Ntripcaster password. Default: %default")
    parser.add_option("-o", "--org", type="string", dest="org", help="Use IBSS and the provided organization for the user. Caster and Port are not needed in this case Default: %default")
    parser.add_option("-b", "--baseorg", type="string", dest="baseorg", help="The org that the base is in. IBSS Only, assumed to be the user org")
    parser.add_option("-t", "--latitude", type="float", dest="lat", default=50.09, help="Your latitude.  Default: %default")
    parser.add_option("-g", "--longitude", type="float", dest="lon", default=8.66, help="Your longitude.  Default: %default")
    parser.add_option("-e", "--height", type="float", dest="height", default=1200, help="Your ellipsoid height.  Default: %default")
    parser.add_option("-s", "--ssl", action="store_true", dest="ssl", default=False, help="Use SSL for the connection")
    parser.add_option("-H", "--host", action="store_true", dest="host", default=False, help="Include host header, should be on for IBSS")
    parser.add_option("-2", "--V2", action="store_true", dest="V2", default=False, help="Make a NTRIP V2 Connection")

    (options, args) = parser.parse_args()
    ntripArgs = {}

    ntripArgs['lat'] = options.lat
    ntripArgs['lon'] = options.lon
    ntripArgs['height'] = options.height
    ntripArgs['host'] = options.host

    if options.ssl:
        import ssl
        ntripArgs['ssl'] = True
    else:
        ntripArgs['ssl'] = False

    if options.org:
        if len(args) != 1:
            print("Incorrect number of arguments for IBSS\n")
            parser.print_help()
            sys.exit(1)
        ntripArgs['user'] = options.user+"." + options.org + ":" + options.password
        if options.baseorg:
            ntripArgs['caster'] = options.baseorg + ".ibss.trimbleos.com"
        else:
            ntripArgs['caster'] = options.org + ".ibss.trimbleos.com"
        if options.ssl:
            ntripArgs['port'] = 52101
        else:
            ntripArgs['port'] = 2101
        ntripArgs['mountpoint'] = args[0]
    else:
        if len(args) != 3:
            print("Incorrect number of arguments for NTRIP\n")
            parser.print_help()
            sys.exit(1)
        ntripArgs['user'] = options.user+":"+options.password
        ntripArgs['caster'] = args[0]
        ntripArgs['port'] = int(args[1])
        ntripArgs['mountpoint'] = args[2]

    ntripArgs['V2'] = options.V2

    n = NtripClient(**ntripArgs)
    n.readLoop()



"""
use vicon data to provide VISION_POSITION_ESTIMATE and GPS_INPUT data
"""

import math
import threading
import time

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import LowPassFilter2p
from pymavlink.rotmat import Vector3
from pymavlink.quaternion import Quaternion
from pymavlink import mavutil
from pymavlink import mavextra

from pyvicon import pyvicon


def get_gps_time(tnow):
    """return gps_week and gps_week_ms for current time"""
    leapseconds = 18
    SEC_PER_WEEK = 7 * 86400

    epoch = 86400*(10*365 + (1980-1969)/4 + 1 + 6 - 2) - leapseconds
    epoch_seconds = int(tnow - epoch)
    week = int(epoch_seconds) // SEC_PER_WEEK
    t_ms = int(tnow * 1000) % 1000
    week_ms = (epoch_seconds % SEC_PER_WEEK) * 1000 + ((t_ms//200) * 200)
    return week, week_ms


class ViconModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(ViconModule, self).__init__(mpstate, "vicon", "vicon", public=False)
        self.console.set_status('VPos', 'VPos -- -- --', row=5)
        self.console.set_status('VAtt', 'VAtt -- -- --', row=5)
        self.vicon_settings = mp_settings.MPSettings(
            [('host', str, "vicon"),
             ('origin_lat', float, -35.363261),
             ('origin_lon', float, 149.165230),
             ('origin_alt', float, 584.0),
             ('vision_rate', int, 14),
             ('vel_filter_hz', float, 30.0),
             ('gps_rate', int, 5),
             ('gps_nsats', float, 16),
             ('object_name', str, None)
             ])
        self.add_command('vicon', self.cmd_vicon, 'VICON control',
                         ["<start>",
                          "<stop>",
                          "set (VICONSETTING)"])
        self.add_completion_function('(VICONSETTING)',
                                     self.vicon_settings.completion)
        self.vicon = None
        self.thread = threading.Thread(target=self.thread_loop)
        self.thread.start()
        self.pos = None
        self.att = None
        self.frame_count = 0
        self.gps_count = 0
        self.vision_count = 0
        self.last_frame_count = 0
        self.vel_filter = LowPassFilter2p.LowPassFilter2p(200.0, 30.0)
        self.actual_frame_rate = 0.0

    def detect_vicon_object(self):
        self.vicon.get_frame()
        object_name = self.vicon_settings.object_name
        if object_name is None:
            # We haven't specified which object we are looking for, so just find the first one
            object_name = self.vicon.get_subject_name(0)
        if object_name is None:
            # No objects found
            return None, None
        segment_name = self.vicon.get_subject_root_segment_name(object_name)
        if segment_name is None:
            # Object we're looking for can't be found
            return None, None
        print("Connected to subject '%s' segment '%s'" % (object_name, segment_name))
        return object_name, segment_name

    def get_vicon_pose(self, object_name, segment_name):

        # get position in mm. Coordinates are in NED
        vicon_pos = self.vicon.get_segment_global_translation(object_name, segment_name)

        if vicon_pos is None:
            # Object is not in view
            return None, None, None, None

        vicon_quat = Quaternion(self.vicon.get_segment_global_quaternion(object_name, segment_name))

        pos_ned = Vector3(vicon_pos * 0.001)
        euler = vicon_quat.euler
        roll, pitch, yaw = euler[0], euler[1], euler[2]
        yaw = math.radians(mavextra.wrap_360(math.degrees(yaw)))

        return pos_ned, roll, pitch, yaw

    def thread_loop(self):
        """background processing"""
        object_name = None
        segment_name = None
        last_pos = None
        last_frame_num = None
        frame_count = 0

        while True:
            if self.vicon is None:
                time.sleep(0.1)
                object_name = None
                continue

            if not object_name:
                object_name, segment_name = self.detect_vicon_object()
                if object_name is None:
                    continue
                last_msg_time = time.time()
                now = time.time()
                last_origin_send = now
                now_ms = int(now * 1000)
                last_gps_send_ms = now_ms
                frame_rate = self.vicon.get_frame_rate()
                frame_dt = 1.0/frame_rate
                last_rate = time.time()
                frame_count = 0
                print("Vicon frame rate %.1f" % frame_rate)

            if self.vicon_settings.gps_rate > 0:
                gps_period_ms = 1000 // self.vicon_settings.gps_rate
            time.sleep(0.01)
            self.vicon.get_frame()
            mav = self.master
            now = time.time()
            now_ms = int(now * 1000)
            frame_num = self.vicon.get_frame_number()

            frame_count += 1
            if now - last_rate > 0.1:
                rate = frame_count / (now - last_rate)
                self.actual_frame_rate = 0.9 * self.actual_frame_rate + 0.1 * rate
                last_rate = now
                frame_count = 0
                self.vel_filter.set_cutoff_frequency(self.actual_frame_rate, self.vicon_settings.vel_filter_hz)

            pos_ned, roll, pitch, yaw = self.get_vicon_pose(object_name, segment_name)
            if pos_ned is None:
                continue
            
            # print(f"XYZ: {pos_ned.x}, {pos_ned.y}, {pos_ned.z}, ")

            if last_frame_num is None or frame_num - last_frame_num > 100 or frame_num <= last_frame_num:
                last_frame_num = frame_num
                last_pos = pos_ned
                continue

            dt = (frame_num - last_frame_num) * frame_dt
            vel = (pos_ned - last_pos) * (1.0/dt)
            last_pos = pos_ned
            last_frame_num = frame_num

            filtered_vel = self.vel_filter.apply(vel)

            if self.vicon_settings.vision_rate > 0:
                dt = now - last_msg_time
                if dt < 1.0 / self.vicon_settings.vision_rate:
                    continue

            last_msg_time = now

            self.pos = pos_ned
            self.att = [math.degrees(roll), math.degrees(pitch), math.degrees(yaw)]
            self.frame_count += 1

            time_us = int(now * 1.0e6)

            if now - last_origin_send > 1 and self.vicon_settings.vision_rate > 0:
                # send a heartbeat msg
                mav.mav.heartbeat_send(mavutil.mavlink.MAV_TYPE_GCS, mavutil.mavlink.MAV_AUTOPILOT_GENERIC, 0, 0, 0)

                # send origin at 1Hz
                mav.mav.set_gps_global_origin_send(self.target_system,
                                                   int(self.vicon_settings.origin_lat*1.0e7),
                                                   int(self.vicon_settings.origin_lon*1.0e7),
                                                   int(self.vicon_settings.origin_alt*1.0e3),
                                                   time_us)
                last_origin_send = now

            if self.vicon_settings.gps_rate > 0 and now_ms - last_gps_send_ms > gps_period_ms:
                '''send GPS data at the specified rate, trying to align on the given period'''
                self.gps_input_send(now, pos_ned, yaw, filtered_vel)
                last_gps_send_ms = (now_ms//gps_period_ms) * gps_period_ms
                self.gps_count += 1

            if self.vicon_settings.vision_rate > 0:
                # send VISION_POSITION_ESTIMATE
                # we force mavlink1 to avoid the covariances which seem to make the packets too large
                # for the mavesp8266 wifi bridge
                mav.mav.global_vision_position_estimate_send(time_us,
                                                             pos_ned.x, pos_ned.y, pos_ned.z,
                                                             roll, pitch, yaw, force_mavlink1=True)
                self.vision_count += 1

    def gps_input_send(self, time, pos_ned, yaw, gps_vel):
        time_us = int(time * 1.0e6)

        gps_lat, gps_lon = mavextra.gps_offset(self.vicon_settings.origin_lat,
                                               self.vicon_settings.origin_lon,
                                               pos_ned.y, pos_ned.x)
        gps_alt = self.vicon_settings.origin_alt - pos_ned.z
        gps_week, gps_week_ms = get_gps_time(time)
        if self.vicon_settings.gps_nsats >= 6:
            fix_type = 3
        else:
            fix_type = 1
        yaw_cd = int(mavextra.wrap_360(math.degrees(yaw)) * 100)
        if yaw_cd == 0:
            # the yaw extension to GPS_INPUT uses 0 as no yaw support
            yaw_cd = 36000
        self.master.mav.gps_input_send(time_us, 0, 0, gps_week_ms, gps_week, fix_type,
                               int(gps_lat * 1.0e7), int(gps_lon * 1.0e7), gps_alt,
                               1.0, 1.0,
                               gps_vel.x, gps_vel.y, gps_vel.z,
                               0.2, 1.0, 1.0,
                               self.vicon_settings.gps_nsats,
                               yaw_cd)

    def cmd_start(self):
        """start vicon"""
        vicon = pyvicon.PyVicon()
        print("Opening Vicon connection to %s" % self.vicon_settings.host)
        vicon.connect(self.vicon_settings.host)
        print("Configuring vicon")
        vicon.set_stream_mode(pyvicon.StreamMode.ClientPull)
        vicon.enable_marker_data()
        vicon.enable_segment_data()
        vicon.enable_unlabeled_marker_data()
        vicon.enable_device_data()

        # Set the axis mapping to the ardupilot convention (North, East, Down)
        vicon.set_axis_mapping(pyvicon.Direction.Forward, pyvicon.Direction.Right, pyvicon.Direction.Down)
        print(vicon.get_axis_mapping())
        print("vicon ready")
        self.vicon = vicon

    def cmd_vicon(self, args):
        """command processing"""
        if len(args) == 0:
            print("Usage: vicon <set|start|stop>")
            return
        if args[0] == "start":
            self.cmd_start()
        if args[0] == "stop":
            self.vicon = None
        elif args[0] == "set":
            self.vicon_settings.command(args[1:])

    def idle_task(self):
        """run on idle"""
        if not self.pos or not self.att or self.frame_count == self.last_frame_count:
            return
        self.last_frame_count = self.frame_count
        self.console.set_status('VPos', 'Vicon: Pos: %.2fN %.2fE %.2fD' % (self.pos.x, self.pos.y, self.pos.z), row=5)
        self.console.set_status('VAtt', ' Att R:%.2f P:%.2f Y:%.2f GPS %u VIS %u RATE %.1f' % (self.att[0], self.att[1], self.att[2],
                                                                                         self.gps_count, self.vision_count,
                                                                                         self.actual_frame_rate), row=5)


def init(mpstate):
    """initialise module"""
    return ViconModule(mpstate)



import functools
import math
from MAVProxy.modules.mavproxy_map import mp_elevation
import numpy as np
import os
import time

from ..lib.wx_loader import wx

from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipBrightness
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipCenter
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipClearLayer
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipDefaultPopup
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipFlightModeLegend
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipGrid
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipHideObject
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipIcon
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipInformation
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipKeyEvent
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipMenuEvent
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipMouseEvent
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipObject
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipObjectSelection
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipPosition
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipRemoveObject
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipThumbnail
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipZoom
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipFollow
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import SlipFollowObject

from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import win_layout
from MAVProxy.modules.lib import icon

from MAVProxy.modules.lib.mp_menu import MPMenuCheckbox
from MAVProxy.modules.lib.mp_menu import MPMenuItem
from MAVProxy.modules.lib.mp_menu import MPMenuRadio
from MAVProxy.modules.lib.mp_menu import MPMenuSeparator
from MAVProxy.modules.lib.mp_menu import MPMenuSubMenu
from MAVProxy.modules.lib.mp_menu import MPMenuTop


class MPSlipMapFrame(wx.Frame):
    """ The main frame of the viewer
    """
    def __init__(self, state):
        wx.Frame.__init__(self, None, wx.ID_ANY, state.title)
        try:
            self.SetIcon(icon.SimpleIcon("MAP").get_ico())
        except Exception:
            pass
        self.state = state
        state.frame = self
        state.grid = True
        state.follow = True
        state.download = True
        state.popup_object = None
        state.popup_latlon = None
        state.popup_started = False
        state.default_popup = None
        state.panel = MPSlipMapPanel(self, state)
        self.last_layout_send = time.time()
        self.Bind(wx.EVT_IDLE, self.on_idle)
        self.Bind(wx.EVT_SIZE, state.panel.on_size)
        self.legend_checkbox_menuitem_added = False
        
        # create the View menu
        self.menu = MPMenuTop([
            MPMenuSubMenu('View', items=[
                MPMenuCheckbox('Follow\tCtrl+F',
                               'Follow Aircraft',
                               'toggleFollow',
                               checked=state.follow),
                MPMenuCheckbox('Grid\tCtrl+G',
                               'Enable Grid',
                               'toggleGrid',
                               checked=state.grid),
                MPMenuItem('Goto\tCtrl+P',
                           'Goto Position',
                           'gotoPosition'),
                MPMenuItem('Brightness +\tCtrl+B',
                           'Increase Brightness',
                           'increaseBrightness'),
                MPMenuItem('Brightness -\tCtrl+Shift+B',
                           'Decrease Brightness',
                           'decreaseBrightness'),
                MPMenuItem('Zoom +\t+',
                           'Zoom In',
                           'zoomIn'),
                MPMenuItem('Zoom -\t-',
                           'Zoom Out',
                           'zoomOut'),
                MPMenuCheckbox('Download Tiles\tCtrl+D',
                               'Enable Tile Download',
                               'toggleDownload',
                               checked=state.download),
                MPMenuRadio('Service', 'Select map service',
                            returnkey='setService',
                            selected=state.mt.get_service(),
                            items=state.mt.get_service_list())])])
        self.SetMenuBar(self.menu.wx_menu())
        self.Bind(wx.EVT_MENU, self.on_menu)

    def on_menu(self, event):
        '''handle menu selection'''
        state = self.state
        # see if it is a popup menu
        if state.popup_object is not None:
            obj = state.popup_object
            ret = obj.popup_menu.find_selected(event)
            if ret is not None:
                ret.call_handler()
                state.event_queue.put(SlipMenuEvent(state.popup_latlon, event,
                                                    [SlipObjectSelection(obj.key, 0, obj.layer, obj.selection_info())],
                                                    ret))
                state.popup_object = None
                state.popup_latlon = None
        if state.default_popup is not None:
            ret = state.default_popup.popup.find_selected(event)
            if ret is not None:
                ret.call_handler()
                state.event_queue.put(SlipMenuEvent(state.popup_latlon, event, [], ret))

        # otherwise a normal menu
        ret = self.menu.find_selected(event)
        if ret is None:
            return
        ret.call_handler()
        if ret.returnkey == 'toggleGrid':
            state.grid = ret.IsChecked()
        elif ret.returnkey == 'toggleLegend':
            state.legend = ret.IsChecked()
        elif ret.returnkey == 'toggleFollow':
            state.follow = ret.IsChecked()
        elif ret.returnkey == 'toggleDownload':
            state.download = ret.IsChecked()
        elif ret.returnkey == 'setService':
            state.mt.set_service(ret.get_choice())
        elif ret.returnkey == 'gotoPosition':
            state.panel.enter_position()
        elif ret.returnkey == 'increaseBrightness':
            state.brightness += 20
            if state.brightness > 255:
                state.brightness = 255
        elif ret.returnkey == 'decreaseBrightness':
            state.brightness -= 20
            if state.brightness < -255:
                state.brightness = -255
        elif ret.returnkey == 'zoomIn':
            state.panel.change_zoom(1.0/1.2)
        elif ret.returnkey == 'zoomOut':
            state.panel.change_zoom(1.2)
        state.need_redraw = True

    def find_object(self, key, layers):
        '''find an object to be modified'''
        state = self.state

        if layers is None or layers == '':
            layers = state.layers.keys()
        for layer in layers:
            if key in state.layers[layer]:
                return state.layers[layer][key]
        return None

    def follow(self, object):
        '''follow an object on the map'''
        state = self.state
        (px,py) = state.panel.pixmapper(object.latlon)
        ratio = 0.25
        if (px > ratio*state.width and
            px < (1.0-ratio)*state.width and
            py > ratio*state.height and
            py < (1.0-ratio)*state.height):
            # we're in the mid part of the map already, don't move
            return

        if not state.follow:
            # the user has disabled following
            return

        (lat, lon) = object.latlon
        state.panel.re_center(state.width/2, state.height/2, lat, lon)

    def add_legend_checkbox_menuitem(self):
        self.menu.add_to_submenu(['View'], [MPMenuCheckbox('Legend\tCtrl+L',
                                                         'Enable Legend',
                                                         'toggleLegend',
                                                         checked=self.state.legend)
        ])

    def add_object(self, obj):
        '''add an object to a layer'''
        state = self.state
        if not obj.layer in state.layers:
            # its a new layer
            state.layers[obj.layer] = {}
        state.layers[obj.layer][obj.key] = obj
        state.need_redraw = True
        if (not self.legend_checkbox_menuitem_added and
            isinstance(obj, SlipFlightModeLegend)):
            self.add_legend_checkbox_menuitem()
            self.legend_checkbox_menuitem_added = True
            self.SetMenuBar(self.menu.wx_menu())

    def remove_object(self, key):
        '''remove an object by key from all layers'''
        state = self.state
        for layer in state.layers:
            state.layers[layer].pop(key, None)
        state.need_redraw = True

    def on_idle(self, event):
        '''prevent the main loop spinning too fast'''
        state = self.state

        if state.close_window.acquire(False):
            self.state.app.ExitMainLoop()

        now = time.time()
        if now - self.last_layout_send > 1:
            self.last_layout_send = now
            state.event_queue.put(win_layout.get_wx_window_layout(self))

        # receive any display objects from the parent
        obj = None

        while not state.object_queue.empty():
            obj = state.object_queue.get()

            if isinstance(obj, win_layout.WinLayout):
                win_layout.set_wx_window_layout(self, obj)
                
            if isinstance(obj, SlipObject):
                self.add_object(obj)

            if isinstance(obj, SlipPosition):
                # move an object
                object = self.find_object(obj.key, obj.layer)
                if object is not None:
                    object.update_position(obj)
                    if getattr(object, 'follow', False):
                        self.follow(object)
                    if obj.label is not None:
                        object.label = obj.label
                    if obj.colour is not None:
                        object.colour = obj.colour
                    state.need_redraw = True

            if isinstance(obj, SlipDefaultPopup):
                state.default_popup = obj

            if isinstance(obj, SlipInformation):
                # see if its a existing one or a new one
                if obj.key in state.info:
#                    print('update %s' % str(obj.key))
                    state.info[obj.key].update(obj)
                else:
#                    print('add %s' % str(obj.key))
                    state.info[obj.key] = obj
                state.need_redraw = True

            if isinstance(obj, SlipCenter):
                # move center
                (lat,lon) = obj.latlon
                state.panel.re_center(state.width/2, state.height/2, lat, lon)
                state.need_redraw = True

            if isinstance(obj, SlipZoom):
                # change zoom
                state.panel.set_ground_width(obj.ground_width)
                state.need_redraw = True
                
            if isinstance(obj, SlipFollow):
                # enable/disable follow
                state.follow = obj.enable

            if isinstance(obj, SlipFollowObject):
                # enable/disable follow on an object
                for layer in state.layers:
                    if obj.key in state.layers[layer]:
                        if hasattr(state.layers[layer][obj.key], 'follow'):
                            state.layers[layer][obj.key].follow = obj.enable
                
            if isinstance(obj, SlipBrightness):
                # set map brightness
                state.brightness = obj.brightness
                state.need_redraw = True

            if isinstance(obj, SlipClearLayer):
                # remove all objects from a layer
                if obj.layer in state.layers:
                    state.layers.pop(obj.layer)
                state.need_redraw = True

            if isinstance(obj, SlipRemoveObject):
                # remove an object by key
                for layer in state.layers:
                    if obj.key in state.layers[layer]:
                        state.layers[layer].pop(obj.key)
                state.need_redraw = True

            if isinstance(obj, SlipHideObject):
                # hide an object by key
                for layer in state.layers:
                    if obj.key in state.layers[layer]:
                        state.layers[layer][obj.key].set_hidden(obj.hide)
                state.need_redraw = True

        if state.timelim_pipe is not None:
            while state.timelim_pipe[1].poll():
                try:
                    obj = state.timelim_pipe[1].recv()
                except Exception:
                    state.timelim_pipe = None
                    break
                for layer in state.layers:
                    for key in state.layers[layer].keys():
                        state.layers[layer][key].set_time_range(obj)
                state.need_redraw = True

        if obj is None:
            time.sleep(0.05)


class MPSlipMapPanel(wx.Panel):
    """ The image panel
    """
    def __init__(self, parent, state):
        from MAVProxy.modules.lib import mp_widgets

        wx.Panel.__init__(self, parent)
        self.state = state
        self.img = None
        self.map_img = None
        self.redraw_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_redraw_timer, self.redraw_timer)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        self.redraw_timer.Start(200)
        self.mouse_pos = None
        self.mouse_down = None
        self.click_pos = None
        self.last_click_pos = None
        if state.elevation:
            self.ElevationMap = mp_elevation.ElevationModel()

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.mainSizer)

        # display for lat/lon/elevation
        self.position = wx.TextCtrl(self, style=wx.TE_MULTILINE|wx.TE_READONLY)
        if os.name == 'nt':
            self.position.SetValue("line 1\nline 2\n")
            size = self.position.GetBestSize()
            self.position.SetMinSize(size)
            self.position.SetValue("")
        else:
            textsize = tuple(self.position.GetFullTextExtent('line 1\nline 2\n')[0:2])
            self.position.SetMinSize(textsize)

        self.mainSizer.AddSpacer(2)
        self.mainSizer.Add(self.position, flag=wx.LEFT | wx.BOTTOM | wx.GROW, border=0)
        self.position.Bind(wx.EVT_SET_FOCUS, self.on_focus)

        # a place to put control flags
        self.controls = wx.BoxSizer(wx.HORIZONTAL)
        self.mainSizer.Add(self.controls, 0, flag=wx.ALIGN_LEFT | wx.TOP | wx.GROW)
        self.mainSizer.AddSpacer(2)

        # a place to put information like image details
        self.information = wx.BoxSizer(wx.HORIZONTAL)
        self.mainSizer.Add(self.information, 0, flag=wx.ALIGN_LEFT | wx.TOP | wx.GROW)
        self.mainSizer.AddSpacer(2)

        # panel for the main map image
        self.imagePanel = mp_widgets.ImagePanel(self, np.zeros((state.height, state.width, 3), dtype=np.uint8))
        self.mainSizer.Add(self.imagePanel, flag=wx.GROW, border=5)
        self.imagePanel.Bind(wx.EVT_MOUSE_EVENTS, self.on_mouse)
        self.imagePanel.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.imagePanel.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)

        # a function to convert from (lat,lon) to (px,py) on the map
        self.pixmapper = functools.partial(self.pixel_coords)

        self.last_view = None
        self.redraw_map()
        state.frame.Fit()

    def on_focus(self, event):
        '''called when the panel gets focus'''
        self.imagePanel.SetFocus()

    def current_view(self):
        '''return a tuple representing the current view'''
        state = self.state
        return (state.lat, state.lon, state.width, state.height,
                state.ground_width, state.mt.tiles_pending())

    def coordinates(self, x, y):
        '''return coordinates of a pixel in the map'''
        state = self.state
        return state.mt.coord_from_area(x, y, state.lat, state.lon, state.width, state.ground_width)

    def constrain_latlon(self):
        self.state.lat = mp_util.constrain(self.state.lat, -85, 85)
        self.state.lon = mp_util.wrap_180(self.state.lon)

    def re_center(self, x, y, lat, lon):
        '''re-center view for pixel x,y'''
        state = self.state
        if lat is None or lon is None:
            return
        (lat2,lon2) = self.coordinates(x, y)
        distance = mp_util.gps_distance(lat2, lon2, lat, lon)
        bearing  = mp_util.gps_bearing(lat2, lon2, lat, lon)
        (state.lat, state.lon) = mp_util.gps_newpos(state.lat, state.lon, bearing, distance)
        self.constrain_latlon()

    def set_ground_width(self, ground_width):
        '''set ground width of view'''
        state = self.state
        state.ground_width = ground_width
        state.panel.re_center(state.width/2, state.height/2, state.lat, state.lon)
         
    def change_zoom(self, zoom):
        '''zoom in or out by zoom factor, keeping centered'''
        state = self.state
        if self.mouse_pos:
            (x,y) = (self.mouse_pos.x, self.mouse_pos.y)
        else:
            (x,y) = (state.width/2, state.height/2)
        (lat,lon) = self.coordinates(x, y)
        lat = mp_util.constrain(lat, -85, 85)
        lon = mp_util.constrain(lon, -180, 180)
        state.ground_width *= zoom
        # limit ground_width to sane values
        state.ground_width = max(state.ground_width, 2)
        state.ground_width = min(state.ground_width, 20000000)
        self.re_center(x,y, lat, lon)

    def enter_position(self):
        '''enter new position'''
        state = self.state
        dlg = wx.TextEntryDialog(self, 'Enter new position', 'Position')
        dlg.SetValue("%f %f" % (state.lat, state.lon))
        if dlg.ShowModal() == wx.ID_OK:
            latlon = dlg.GetValue().split()
            dlg.Destroy()
            state.lat = float(latlon[0])
            state.lon = float(latlon[1])
            self.constrain_latlon()
            self.re_center(state.width/2,state.height/2, state.lat, state.lon)
            self.redraw_map()

    def update_position(self):
        '''update position text'''
        state = self.state
        pos = self.mouse_pos
        newtext = ''
        alt = 0
        if pos is not None:
            (lat,lon) = self.coordinates(pos.x, pos.y)
            newtext += 'Cursor: %.8f %.8f (%s)' % (lat, lon, mp_util.latlon_to_grid((lat, lon)))
            if state.elevation:
                alt = self.ElevationMap.GetElevation(lat, lon)
                if alt is not None:
                    newtext += ' %.1fm %uft' % (alt, alt*3.28084)
        state.mt.set_download(state.download)
        pending = 0
        if state.download:
            pending = state.mt.tiles_pending()
        if pending:
            newtext += ' Map Downloading %u ' % pending
        if alt == -1:
            newtext += ' SRTM Downloading '
        newtext += '\n'
        if self.click_pos is not None:
            newtext += 'Click: %.8f %.8f (%s %s) (%s)' % (self.click_pos[0], self.click_pos[1],
                                                      mp_util.degrees_to_dms(self.click_pos[0]),
                                                      mp_util.degrees_to_dms(self.click_pos[1]),
                                                      mp_util.latlon_to_grid(self.click_pos))
        if self.last_click_pos is not None:
            distance = mp_util.gps_distance(self.last_click_pos[0], self.last_click_pos[1],
                                            self.click_pos[0], self.click_pos[1])
            bearing = mp_util.gps_bearing(self.last_click_pos[0], self.last_click_pos[1],
                                            self.click_pos[0], self.click_pos[1])
            newtext += '  Distance: %.3fm %.3fnm Bearing %.1f' % (distance, distance*0.000539957, bearing)
        if newtext != state.oldtext:
            self.position.Clear()
            self.position.WriteText(newtext)
            state.oldtext = newtext

    def pixel_coords(self, latlon, reverse=False):
        '''return pixel coordinates in the map image for a (lat,lon)
        if reverse is set, then return lat/lon for a pixel coordinate
        '''
        state = self.state
        if reverse:
            (x,y) = latlon
            return self.coordinates(x,y)
        (lat,lon) = (latlon[0], latlon[1])
        return state.mt.coord_to_pixel(state.lat, state.lon, state.width, state.ground_width, lat, lon)

    def draw_objects(self, objects, bounds, img):
        '''draw objects on the image'''
        keys = sorted(objects.keys())
        for k in keys:
            obj = objects[k]
            if not self.state.legend and isinstance(obj, SlipFlightModeLegend):
                continue
            bounds2 = obj.bounds()
            if bounds2 is None or mp_util.bounds_overlap(bounds, bounds2):
                obj.draw(img, self.pixmapper, bounds)

    def redraw_map(self):
        '''redraw the map with current settings'''
        state = self.state

        view_same = (self.last_view is not None and self.map_img is not None and self.last_view == self.current_view())

        if view_same and not state.need_redraw:
            return

        # get the new map
        self.map_img = state.mt.area_to_image(state.lat, state.lon,
                                              state.width, state.height, state.ground_width)
        if state.brightness != 0: # valid state.brightness range is [-255, 255]
            brightness = np.uint8(np.abs(state.brightness))
            if state.brightness > 0:
                self.map_img = np.where((255 - self.map_img) < brightness, 255, self.map_img + brightness)
            else:
                self.map_img = np.where((255 + self.map_img) < brightness, 0, self.map_img - brightness)
            
        # find display bounding box
        (lat2,lon2) = self.coordinates(state.width-1, state.height-1)
        bounds = (lat2, state.lon, state.lat-lat2, mp_util.wrap_180(lon2-state.lon))

        # get the image
        img = self.map_img.copy()

        # possibly draw a grid
        if state.grid:
            SlipGrid('grid', layer=3, linewidth=1, colour=(255,255,0)).draw(img, self.pixmapper, bounds)

        # draw layer objects
        keys = state.layers.keys()
        keys = sorted(list(keys))
        for k in keys:
            self.draw_objects(state.layers[k], bounds, img)

        # draw information objects
        for key in state.info:
            state.info[key].draw(state.panel, state.panel.information)

        # display the image
        self.imagePanel.set_image(img)

        self.update_position()

        self.mainSizer.Fit(self)
        self.Refresh()
        self.last_view = self.current_view()
        self.SetFocus()
        state.need_redraw = False

    def on_redraw_timer(self, event):
        '''the redraw timer ensures we show new map tiles as they
        are downloaded'''
        self.redraw_map()

    def on_size(self, event):
        '''handle window size changes'''
        state = self.state
        size = event.GetSize()
        state.width = size.width
        state.height = size.height
        self.redraw_map()

    def on_mouse_wheel(self, event):
        '''handle mouse wheel zoom changes'''
        # >>> print -1/120
        # -1
        wheel_rotation = event.GetWheelRotation()
        rotation = abs(wheel_rotation) // event.GetWheelDelta()
        if rotation == 0:
            return
        zoom = 1.1 * rotation
        if wheel_rotation > 0:
            # zooming out
            zoom = 1.0/zoom
        self.change_zoom(zoom)
        self.redraw_map()

    def selected_objects(self, pos):
        '''return a list of matching objects for a position'''
        state = self.state
        selected = []
        (px, py) = pos
        for layer in state.layers:
            for key in state.layers[layer]:
                obj = state.layers[layer][key]
                distance = obj.clicked(px, py)
                if distance is not None:
                    selected.append(SlipObjectSelection(key, distance, layer, extra_info=obj.selection_info()))
        selected.sort(key=lambda c: c.distance)
        return selected

    def show_popup(self, selected, pos):
        '''show popup menu for an object'''
        state = self.state
        if selected.popup_menu is not None:
            import copy
            popup_menu = selected.popup_menu
            if state.default_popup is not None and state.default_popup.combine:
                popup_menu = copy.deepcopy(popup_menu)
                popup_menu.add(MPMenuSeparator())
                popup_menu.combine(state.default_popup.popup)
            wx_menu = popup_menu.wx_menu()
            state.frame.PopupMenu(wx_menu, pos)

    def show_default_popup(self, pos):
        '''show default popup menu'''
        state = self.state
        if state.default_popup.popup is not None:
            wx_menu = state.default_popup.popup.wx_menu()
            state.frame.PopupMenu(wx_menu, pos)

    def on_mouse(self, event):
        '''handle mouse events'''
        state = self.state
        pos = event.GetPosition()
        if event.Leaving():
            self.mouse_pos = None
        else:
            self.mouse_pos = pos
        self.update_position()

        if hasattr(event, 'ButtonIsDown'):
            any_button_down = event.ButtonIsDown(wx.MOUSE_BTN_ANY)
            left_button_down = event.ButtonIsDown(wx.MOUSE_BTN_LEFT)
            right_button_down = event.ButtonIsDown(wx.MOUSE_BTN_RIGHT)
        else:
            left_button_down = event.leftIsDown
            right_button_down = event.rightIsDown
            any_button_down = left_button_down or right_button_down

        if any_button_down or event.ButtonUp():
            # send any event with a mouse button to the parent
            latlon = self.coordinates(pos.x, pos.y)
            selected = self.selected_objects(pos)
            state.event_queue.put(SlipMouseEvent(latlon, event, selected))
            if event.RightDown():
                state.popup_object = None
                state.popup_latlon = None
                if len(selected) > 0:
                    obj = state.layers[selected[0].layer][selected[0].objkey]
                    if obj.popup_menu is not None:
                        state.popup_object = obj
                        state.popup_latlon = latlon
                        self.show_popup(obj, pos)
                        state.popup_started = True
                if not state.popup_started and state.default_popup is not None:
                    state.popup_latlon = latlon
                    self.show_default_popup(pos)
                    state.popup_started = True

        if not right_button_down:
            state.popup_started = False

        if event.LeftDown() or event.RightDown():
            self.mouse_down = pos
            self.last_click_pos = self.click_pos
            self.click_pos = self.coordinates(pos.x, pos.y)

        if event.Dragging() and left_button_down:
            # drag map to new position
            newpos = pos
            if self.mouse_down and newpos:
                dx = (self.mouse_down.x - newpos.x)
                dy = -(self.mouse_down.y - newpos.y)
                pdist = math.sqrt(dx**2 + dy**2)
                if pdist > state.drag_step:
                    bearing = math.degrees(math.atan2(dx, dy))
                    distance = (state.ground_width/float(state.width)) * pdist
                    newlatlon = mp_util.gps_newpos(state.lat, state.lon, bearing, distance)
                    (state.lat, state.lon) = newlatlon
                    self.constrain_latlon()
                    self.mouse_down = newpos
                    self.redraw_map()

    def clear_thumbnails(self):
        '''clear all thumbnails from the map'''
        state = self.state
        for l in state.layers:
            keys = state.layers[l].keys()[:]
            for key in keys:
                if (isinstance(state.layers[l][key], SlipThumbnail)
                    and not isinstance(state.layers[l][key], SlipIcon)):
                    state.layers[l].pop(key)

    def on_key_down(self, event):
        '''handle keyboard input'''
        state = self.state

        # send all key events to the parent
        if self.mouse_pos:
            latlon = self.coordinates(self.mouse_pos.x, self.mouse_pos.y)
            selected = self.selected_objects(self.mouse_pos)
            state.event_queue.put(SlipKeyEvent(latlon, event, selected))

        if hasattr(event,'GetUnicodeKey'):
            c = event.GetUnicodeKey()
        else:
            c = event.GetUniChar()
        if c == ord('+') or (c == ord('=') and event.ShiftDown()):
            self.change_zoom(1.0/1.2)
        elif c == ord('-'):
            self.change_zoom(1.2)
        elif c == ord('G') and not event.ControlDown():
            self.enter_position()
        elif c == ord('C'):
            self.clear_thumbnails()
        else:
            # propogate event:
            event.Skip()



# Copyright (C) 2016  Intel Corporation. All rights reserved.
#
# This file is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import division

import math
import os.path
import time

from OpenGL.GL import *
from pymavlink import quaternion
from pymavlink.rotmat import Vector3

from MAVProxy.modules import mavproxy_magical as magical
from MAVProxy.modules.lib import geodesic_grid
from MAVProxy.modules.lib import opengl
from MAVProxy.modules.lib import wavefront as wv
from MAVProxy.modules.lib.mp_util import quaternion_to_axis_angle
from MAVProxy.modules.lib.wx_loader import wx
from MAVProxy.modules.mavproxy_magical import glrenderer

class Renderer(glrenderer.Renderer):
    def __init__(self, background):
        super(Renderer, self).__init__(background)

        glEnable(GL_DEPTH_TEST)
        glEnable(GL_MULTISAMPLE)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        self.visible = [False for _ in range(len(geodesic_grid.sections))]

        self.common_model_transform = opengl.Transform()

        self.last_render_quaternion = quaternion.Quaternion(
                self.common_model_transform.quaternion)

        vertices = [(p.x, p.y, p.z) for t in geodesic_grid.sections for p in t]
        self.sphere = opengl.Object(vertices, enable_alpha=True)
        self.sphere.local.scale(0.46)
        self.sphere.model = self.common_model_transform

        self.base_color = Vector3(0.08, 0.68, 0.706667)

        self.vehicle = None

        path = os.path.join(magical.datapath, 'arrow.obj')
        obj = wv.ObjParser(filename=path).parse()
        self.mag = opengl.WavefrontObject(obj)
        self.mag.local.scale(.88)

    def get_load_progress(self):
        return self.vehicle_loader.get_progress()

    def rotate(self, vector, angle, rotate_mag=False):
        self.common_model_transform.rotate(vector, angle)
        if rotate_mag and self.mag:
            self.mag.model.rotate(vector, angle)

    def set_attitude(self, roll, pitch, yaw):
        self.common_model_transform.set_euler(roll, pitch, yaw)

    def set_mag(self, mag):
        if not mag.length():
            return
        m = self.common_model_transform.apply(mag).normalized()
        axis = Vector3(0, 0, 1) % m
        angle = math.acos(Vector3(0, 0, 1) * m)
        self.mag.model.set_rotation(axis, angle)

    def render(self):
        super(Renderer, self).render()

        if self.vehicle:
            self.vehicle.draw(self.program)

        self.sphere.material.set_color(1.2 * self.base_color)
        self.sphere.material.specular_exponent = 4
        self.sphere.material.alpha = 1.0
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        self.sphere.draw(self.program)

        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        self.mag.draw(self.program)

        self.sphere.material.set_color(self.base_color)
        self.sphere.material.alpha = .4
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        faces = [i for i in range(80) if self.visible[i]]
        self.sphere.draw(self.program, faces=faces, camera=self.camera)

        self.last_render_quaternion = quaternion.Quaternion(
                self.common_model_transform.quaternion)

    def set_section_visible(self, section, visible=True):
        self.visible[section] = visible

    def clear_sections(self):
        for i in range(len(self.visible)):
            self.visible[i] = False

    def set_vehicle_wavefront(self, vehicle):
        self.vehicle = opengl.WavefrontObject(vehicle)
        # FIXME: this class shouldn't need to be aware of the proper scaling
        self.vehicle.local.scale(3.5)
        self.vehicle.model = self.common_model_transform

class GeodesicGrid(glrenderer.GLCanvas):
    def __init__(self, *k, **kw):
        super(GeodesicGrid, self).__init__(*k, **kw)

        self.vehicle_wavefront = None
        self.dragging = False

        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(wx.EVT_MOTION, self.OnMotion)

        self.attitude_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnAttitudeTimer, self.attitude_timer)
        self.attitude_timer.Start(40)

        self.gyro = Vector3(0, 0, 0)
        self.mag = Vector3(0, 0, 0)
        self.filtered_mag = Vector3(0, 0, 0)
        self.attitude_timer_last = 0
        self.attitude_timestamp = 0

    def CreateRenderer(self):
        r, g, b = self.GetParent().GetBackgroundColour()
        self.renderer = Renderer(
            background=(r / 255.0, g / 255.0, b / 255.0, 1),
        )
        if self.vehicle_wavefront:
            self.renderer.set_vehicle_wavefront(self.vehicle_wavefront)
            self.vehicle_wavefront = None

    def SetVehicleWavefront(self, vehicle):
        if self.renderer:
            self.SetCurrent(self.context)
            self.renderer.set_vehicle_wavefront(vehicle)
            self.Refresh()
            return
        self.vehicle_wavefront = vehicle

    def SetMag(self, x, y, z):
        self.mag = Vector3(x, y, z)

    def SetAttitude(self, roll, pitch, yaw, timestamp):
        if not self.renderer:
            return

        dt = 0xFFFFFFFF & (timestamp - self.attitude_timestamp)
        dt *= 0.001
        self.attitude_timestamp = timestamp

        desired_quaternion = quaternion.Quaternion((roll, pitch, yaw))
        desired_quaternion.normalize()
        error = desired_quaternion / self.renderer.common_model_transform.quaternion
        error.normalize()
        self.gyro = quaternion_to_axis_angle(error) * (1.0 / dt)

    def OnAttitudeTimer(self, evt):
        if not self.renderer:
            return

        if self.dragging:
            return

        t = time.time()
        dt = t - self.attitude_timer_last
        self.attitude_timer_last = t

        angle = self.gyro.length()
        angle *= dt
        self.renderer.rotate(self.gyro, angle)

        alpha = 0.8
        self.filtered_mag = alpha * self.filtered_mag + (1 - alpha) * self.mag
        self.renderer.set_mag(self.filtered_mag)

        q = self.renderer.common_model_transform.quaternion
        diff = q / self.renderer.last_render_quaternion
        angle = quaternion_to_axis_angle(diff).length()
        if angle < math.radians(.5):
            # Save some CPU time
            return

        self.Refresh()

    def CalcRotationVector(self, dx, dy):
        angle = math.degrees(math.atan2(dy, dx))
        # Make angle discrete on multiples of 45 degrees
        angle = angle + 22.5 - (angle + 22.5) % 45

        if angle % 90 == 45:
            x, y = 1, 1
        elif (angle / 90) % 2 == 0:
            x, y = 1, 0
        else:
            x, y = 0, 1

        if abs(angle) > 90:
            x *= -1
        if angle < 0:
            y *= -1

        # NED coordinates from the camera point of view
        dx, dy, dz = 0, x, y
        # Get rotation axis by rotating the moving vector -90 degrees on x axis
        self.rotation_vector = Vector3(dx, dz, -dy)

    def GetDeltaAngle(self):
        radius = 60.0
        pos = wx.GetMousePosition()
        d = pos - self.motion_reference
        self.motion_reference = pos

        self.CalcRotationVector(d.x, d.y)

        arc = math.sqrt(d.x**2 + d.y**2)
        return arc / radius

    def OnLeftDown(self, evt):
        self.motion_reference = wx.GetMousePosition()
        self.dragging = True

    def OnLeftUp(self, evt):
        self.dragging = False

    def OnMotion(self, evt):
        if hasattr(evt, 'ButtonIsDown'):
            left_button_down = evt.ButtonIsDown(wx.MOUSE_BTN_LEFT)
        else:
            left_button_down = evt.leftIsDown
        if not evt.Dragging() or not left_button_down:
            return

        angle = self.GetDeltaAngle()
        self.renderer.rotate(self.rotation_vector, angle, rotate_mag=True)

        self.Refresh()

    def UpdateVisibleSections(self, visible):
        for i, v in enumerate(visible):
            self.renderer.set_section_visible(i, v)
        self.Refresh()
